# (c) Copyright Riverlane 2020-2025.
"""
This module contains a class CSSStage, which encapsulates the logic
of stage generation for quantum memory experiment using CSS codes.
"""

import itertools
from collections.abc import Iterable, Mapping, Sequence
from functools import cached_property

import stim
from deltakit_circuit import Circuit, Coordinate, GateLayer, Qubit
from deltakit_circuit._basic_maps import BASIS_TO_PAULI
from deltakit_circuit.gates import (
    MPP,
    MX,
    RX,
    I,
    OneQubitCliffordGate,
    OneQubitMeasurementGate,
    OneQubitResetGate,
    TwoOperandGate,
)

from deltakit_explorer.codes._css._detectors import (
    _calculate_detector_coordinates,
    get_between_round_detectors_and_coordinate_shifts,
)
from deltakit_explorer.codes._css._observables import _construct_observables
from deltakit_explorer.codes._css._stabiliser_helper_functions import (
    _get_data_qubits_from_stabilisers,
    _transform_stabiliser,
    get_entangling_layer,
)
from deltakit_explorer.codes._stabiliser import Stabiliser


class CSSStage:
    """
    Class representing a computational stage in a CSS code experiment. Such a
    stage can consist of some initial qubit measurements, some number of
    rounds of syndrome extraction and some final qubit resets. A full
    experiment can then be constructed by putting together several stages.

    Parameters
    ----------
    stabilisers : Optional[Sequence[Sequence[Stabiliser]]]
        The stabilisers to be measured using syndrome extraction. Each sequence
        gives those stabilisers which are to be measured simultaneously, and
        thus each stabiliser within a sequence should have the same length. The
        sequences are in order of measurement. By default, None.
    num_rounds : int, optional
        The number of rounds of syndrome extraction to be performed. By
        default, 0.
    first_round_measurements : Optional[Sequence[Union[MPP, OneQubitMeasurementGate]]]
        Measurement gates to be included before the first round of syndrome
        extraction. By default, None.
    first_round_gates : Optional[Iterable[Union[OneQubitCliffordGate, TwoOperandGate]]]
        Clifford unitary gates to be included before the first round of syndrome extraction.
        The gates can be general as long as

        - no qubit appears in two gates, and
        - each gate acts on at least one data qubit.

        By default, None.
    final_round_resets : Optional[Iterable[OneQubitResetGate]]
        Reset gates to be included after the final round of syndrome extraction.
        By default, None.
    observable_definitions : Optional[Dict[int, Iterable[Union[Qubit, MPP, OneQubitMeasurementGate]]]]
        Definitions of observables to be included after the initial measurements
        and first round of syndrome extraction. The dictionary keys give the
        observable indices and the values give the measurements which should be
        included in the observable. A measurement is specified with a qubit, in
        which case the most recent one-qubit measurement performed upon the qubit
        should be included, or a measurement gate, in which case the outcome of the
        most recent time the measurement gate was performed should be included.
        By default, None.
    use_ancilla_qubits : bool
        Specifies whether the Code will use an ordered sequence of Stabilisers or not.
        If the code is ordered, one ancilla qubit must be provided for each stabiliser.
        If the code is not ordered, no ancilla qubits must be provided.

    Attributes
    ----------
    measurements_as_stabilisers : Tuple[Stabiliser, ...]
        First round of measurements as a Tuple of Stabilisers. For each measurement,
        the measured operator and the qubit used to measure the outcome are given
        by Stabiliser attributes paulis and ancilla_qubit, respectively.
    final_round_resets : Tuple[Stabiliser, ...]
        Final round resets as a Tuple of Stabilisers. For each reset, the reset
        basis and qubit to reset are given by Stabiliser attributes paulis and
        ancilla_qubit, respectively.
    ordered_stabilisers : Tuple[Stabiliser, ...]
        Stabilisers in order of measurement.
    resets_only : bool
        Whether the stage consists only of final round resets.
    allowable_final_stage : bool
        Whether the stage can be used as the final stage in an experiment.
    first_round : Circuit
        The circuit for the first round of syndrome extraction, including
        the first round data qubit measurements and the subsequent observable
        definitions.
    remaining_rounds : Circuit
        The circuit for all but the first round of syndrome extraction.
        This includes detectors as well as the final round resets.
    """

    def __init__(
        self,
        stabilisers: Sequence[Sequence[Stabiliser]] | None = None,
        num_rounds: int = 0,
        first_round_measurements: Sequence[MPP | OneQubitMeasurementGate] | None = None,
        first_round_gates: Iterable[OneQubitCliffordGate | TwoOperandGate] | None = None,
        final_round_resets: Iterable[OneQubitResetGate] | None = None,
        observable_definitions: Mapping[int, Iterable[Qubit | MPP | OneQubitMeasurementGate]] | None = None,
        use_ancilla_qubits: bool | None = None,
    ):
        self._stabilisers = () if stabilisers is None else stabilisers
        self._num_rounds = num_rounds
        self._use_ancilla_qubits = use_ancilla_qubits

        if self._num_rounds < 0:
            msg = "Number of rounds must be non-negative."
            raise ValueError(msg)
        if self._num_rounds > 0 and all(
            len(stabiliser_set) == 0 for stabiliser_set in self._stabilisers
        ):
            msg = "Non-zero number of rounds requires non-zero number of stabilisers."
            raise ValueError(msg)
        if self._num_rounds == 0 and any(
            len(stabiliser_set) > 0 for stabiliser_set in self._stabilisers
        ):
            msg = "Non-zero number of stabilisers requires non-zero number of rounds."
            raise ValueError(msg)

        self._first_round_measurements = (
            ()
            if first_round_measurements is None
            else tuple(first_round_measurements)
        )
        self._first_round_gates = (
            frozenset() if first_round_gates is None else frozenset(first_round_gates)
        )
        self._gate_qubits = self._calculate_gate_qubits(self._first_round_gates)
        self._final_round_resets = (
            frozenset() if final_round_resets is None else frozenset(final_round_resets)
        )
        self._observable_definitions = (
            {} if observable_definitions is None else observable_definitions
        )

        self._check_disjointness(
            self._stabilisers, self._first_round_measurements, self._final_round_resets
        )

        self._first_round_data_qubits = _get_data_qubits_from_stabilisers(
            self._stabilisers
        )

        self._check_first_round_gates(
            stabilisers=self._stabilisers,
            measurements=self._first_round_measurements,
            gates=self._first_round_gates,
            first_round_data_qubits=self._first_round_data_qubits,
        )

        if self._use_ancilla_qubits is None:
            self._use_ancilla_qubits = self._determine_circuit_construction_method(
                self.ordered_stabilisers
            )

        self._detector_coordinates = _calculate_detector_coordinates(
            self.ordered_stabilisers
        )

    @staticmethod
    def _calculate_gate_qubits(
        first_round_gates: frozenset[OneQubitCliffordGate | TwoOperandGate],
    ) -> tuple[Qubit]:
        """
        Calculate the qubits on which the gates act.
        """
        gate_qubits = tuple(
            itertools.chain.from_iterable(gate.qubits for gate in first_round_gates)
        )
        if len(gate_qubits) > len(set(gate_qubits)):
            msg = "Qubits in first_round_gates have to be unique."
            raise ValueError(msg)
        return gate_qubits

    @staticmethod
    def _check_disjointness(
        stabilisers: Sequence[Sequence[Stabiliser]],
        measurements: Sequence[OneQubitMeasurementGate],
        resets: Iterable[OneQubitResetGate],
    ) -> None:
        """
        Check if measurements and resets act on different qubits than stabilisers.
        """
        stabiliser_qubits = set(
            itertools.chain.from_iterable(
                itertools.chain.from_iterable(
                    stabiliser.data_qubits for stabiliser in stabiliser_set
                )
                for stabiliser_set in stabilisers
            )
        )
        measurement_qubits = {
            gate.qubit
            for gate in measurements
            if isinstance(gate, OneQubitMeasurementGate)
        }
        reset_qubits = {gate.qubit for gate in resets}

        if not stabiliser_qubits.isdisjoint(measurement_qubits):
            msg = (
                "Initial measurement qubits and qubits in stabilisers "
                "should be disjoint."
            )
            raise ValueError(msg)

        if not stabiliser_qubits.isdisjoint(reset_qubits):
            msg = "Final reset qubits and qubits in stabilisers should be disjoint."
            raise ValueError(msg)

    @staticmethod
    def _check_first_round_gates(
        stabilisers: Sequence[Sequence[Stabiliser]],
        measurements: Sequence[OneQubitMeasurementGate],
        gates: frozenset[OneQubitCliffordGate | TwoOperandGate],
        first_round_data_qubits: Sequence[Qubit],
    ) -> None:
        """
        Check that, if first_round_gates is not empty, then the following hold:
          1) first_round_measurements is empty,
          2) stabilisers is not empty,
          3) each gate from first_round gates acts on at least one data qubit.
        """
        if len(gates) == 0:
            return
        # Check 1)
        if len(measurements) != 0:
            msg = (
                "If first_round_gates is non-empty, then first_round_measurements has "
                "to be empty."
            )
            raise ValueError(msg)
        # Check 2).
        if all(
            len(simultaneous_stabilisers) == 0
            for simultaneous_stabilisers in stabilisers
        ):
            msg = (
                "The stabilisers parameter is empty, which is not allowed when "
                "first_round_gates is not empty."
            )
            raise ValueError(msg)
        # Check 3).
        for gate in gates:
            if len(set(gate.qubits).intersection(first_round_data_qubits)) == 0:
                msg = (
                    f"The gate {gate} from first_round_gates is not supported on any "
                    "data qubits, which is not allowed."
                )
                raise ValueError(msg)

    @staticmethod
    def _determine_ancilla_uniqueness(
        stabilisers: Sequence[Sequence[Stabiliser]],
    ) -> bool:
        """
        Determine whether all ancilla qubits are unique.

        Returns
        -------
        bool
            True if all ancilla qubits are unique, False otherwise.
        """
        ancilla_qubits = set()
        for stabiliser_set in stabilisers:
            for stabiliser in stabiliser_set:
                if (qubit := stabiliser.ancilla_qubit) in ancilla_qubits:
                    return False
                ancilla_qubits.add(qubit)
        return True

    @staticmethod
    def _determine_circuit_construction_method(
        stabilisers: tuple[Stabiliser, ...],
    ) -> bool:
        """
        Check whether the stabilisers have ancilla qubits defined and thus determine
        whether the full syndrome extraction circuit should be constructed or if
        MPPs should be used.

        Parameters
        ----------
        stabilisers : Tuple[Stabiliser, ...]
            The stabilisers for which we wish to perform syndrome extraction.

        Returns
        -------
        bool
            Whether the full syndrome extraction circuit should be constructed (True)
            or MPPs should be used (False).

        Raises
        ------
        ValueError
            If there is a mix of stabilisers with and without an ancilla qubit
            defined.
        """
        ancillas_defined = [
            isinstance(stabiliser.ancilla_qubit, Qubit) for stabiliser in stabilisers
        ]
        if all(ancillas_defined):
            return True
        if not any(ancillas_defined):
            return False
        msg = "Either all stabilisers or no stabilisers should have an ancilla defined."
        raise ValueError(msg)

    def _construct_mpp_syndrome_extraction_layers(self) -> list[GateLayer]:
        """Construct the syndrome extraction circuit for the stabilisers using MPPs."""
        layers = []
        for stabilisers_set in self._stabilisers:
            if len(stabilisers_set) == 0:
                continue

            layers.append(
                GateLayer({I(qubit) for qubit in self._first_round_data_qubits})
            )

            for stabiliser in stabilisers_set:
                mpp = MPP([pauli for pauli in stabiliser.paulis if pauli is not None])
                # Fill mpp_layers such that no qubit clashes occur
                # Start a new layer if last layer prohibits current qubit
                if any(qubit in layers[-1].qubits for qubit in mpp.qubits):
                    layers.append(GateLayer(mpp))
                else:
                    layers[-1].add_gates(mpp)

        return layers

    def _construct_full_syndrome_extraction_layers(self) -> list[GateLayer]:
        """
        Construct the syndrome extraction circuit for the stabilisers using gates.
        """
        layers: list[GateLayer] = []
        for timestep, timestep_stabs in enumerate(self._stabilisers):
            ancilla_qubits = tuple(stab.ancilla_qubit for stab in timestep_stabs)
            layers.append(GateLayer(RX(q) for q in ancilla_qubits))

            if timestep == 0:
                layers.append(GateLayer(I(q) for q in self._first_round_data_qubits))

            num_layers = len(timestep_stabs[0].paulis)
            layers.extend(
                get_entangling_layer(timestep_stabs, i) for i in range(num_layers)
            )

            layers.append(GateLayer(MX(q) for q in ancilla_qubits))

        return layers

    def _construct_syndrome_extraction_layers(self) -> list[GateLayer]:
        """
        Construct the layers for the syndrome extraction circuit.
        """
        if self._use_ancilla_qubits:
            return self._construct_full_syndrome_extraction_layers()
        return self._construct_mpp_syndrome_extraction_layers()

    @property
    def detector_coordinates(self) -> tuple[Coordinate, ...]:
        """
        Tuple of coordinates of detectors, computed with respect
        to stabilisers.

        Returns
        -------
        Tuple[Coordinate, ...]
            Tuple of detector coordinates.
        """
        return self._detector_coordinates

    @cached_property
    def measurements_as_stabilisers(self) -> tuple[Stabiliser, ...]:
        """
        Tuple of stabiliser objects representing measurements.

        Returns
        -------
        Tuple[Stabiliser, ...]
            Tuple of stabiliser objects.
        """
        stabs = []
        for gate in self._first_round_measurements:
            if isinstance(gate, OneQubitMeasurementGate):
                stabs.append(
                    Stabiliser(
                        paulis=(BASIS_TO_PAULI[gate.basis](gate.qubit),),
                        ancilla_qubit=gate.qubit,
                    )
                )
            else:
                stabs.append(Stabiliser(paulis=gate.pauli_product))
        return tuple(stabs)

    @cached_property
    def resets_as_stabilisers(self) -> tuple[Stabiliser, ...]:
        """
        Tuple of stabiliser objects representing final round resets.

        Returns
        -------
        Tuple[Stabiliser, ...]
            Tuple of stabiliser objects.
        """
        return tuple(
            Stabiliser(
                paulis=(BASIS_TO_PAULI[gate.basis](gate.qubit),),
                ancilla_qubit=gate.qubit,
            )
            for gate in self._final_round_resets
        )

    @cached_property
    def ordered_stabilisers(self) -> tuple[Stabiliser, ...]:
        """
        Flattened version of the stabilisers we measure in the last round of the stage,
        in order of measurement. This is constructed straight from self._stabilisers.
        """
        return tuple(
            itertools.chain.from_iterable(
                stabiliser_set for stabiliser_set in self._stabilisers
            )
        )

    @cached_property
    def stabilisers_before(self) -> tuple[Stabiliser, ...]:
        """
        A transformed version of ordered_stabilisers with removed ancilla_qubits that
        correspond to the stabilisers before first_round_gates. As ancillas are changed
        to Nones, this should be used only for operator_repr comparison.

        Returns
        -------
        Tuple[Stabiliser, ...]
            The stabilisers before applying the gates in self.first_round_gates.
        """
        data_qubits = tuple(self._first_round_data_qubits)
        tableau_qubits: tuple[Qubit] = data_qubits + self._gate_qubits

        # Create tableau that implements _first_round_gates
        tableau = stim.Tableau(len(tableau_qubits))
        for gate in self._first_round_gates:
            gate_tableau = stim.Tableau.from_named_gate(gate.stim_string)
            qubits = list(gate.qubits)
            tableau.append(
                gate=gate_tableau,
                targets=[tableau_qubits.index(qubit) for qubit in qubits],
            )
        inverse_tableau = tableau.inverse()

        return tuple(
            _transform_stabiliser(stab, tableau_qubits, inverse_tableau)
            for stab in self.ordered_stabilisers
        )

    @cached_property
    def resets_only(self) -> bool:
        """
        True if none of ordered stabilisers, first round measurements,
        and observable definitions are present.

        Returns
        -------
        bool
            True if resets only.
        """
        return not (
            len(self.ordered_stabilisers) > 0
            or len(self._first_round_measurements) > 0
            or len(self._observable_definitions) > 0
        )

    @cached_property
    def allowable_final_stage(self) -> bool:
        """
        Return whether a stage is allowable as the final stage in an experiment.
        True if there are no final round resets and one of the following holds:
        - there is no stabiliser measurement
        - stabilisers are measured for one round only without the use of ancillas

        Returns
        -------
        bool
            True if the above properties hold, else False.
        """
        if len(self._final_round_resets) > 0:
            return False
        return not (
            len(self.ordered_stabilisers) > 0
            and (self._num_rounds > 1 or self._use_ancilla_qubits)
        )

    @property
    def first_round(self) -> Circuit:
        """
        Circuit layers of the first stage round.

        Returns
        -------
        Circuit
            A circuit with the first stage round.
        """
        layers: list[GateLayer] = []
        current_layer = GateLayer()
        for gate in list(self._first_round_measurements) + list(
            self._first_round_gates
        ):
            if any(qubit in current_layer.qubits for qubit in gate.qubits):
                layers.append(current_layer)
                current_layer = GateLayer(gate)
            else:
                current_layer.add_gates(gate)
        if len(current_layer.gates) != 0:
            layers.append(current_layer)

        if self._num_rounds >= 1:
            layers += self._construct_syndrome_extraction_layers()

        mmt_gates = [mmt for layer in layers for mmt in layer.measurement_gates]
        layers += _construct_observables(self._observable_definitions, mmt_gates)

        return Circuit(layers)

    @property
    def remaining_rounds(self) -> Circuit:
        """
        The rest of the circuit layers, which do not include
        actions specific to the first layer,
        and attach final rounds resets.

        Returns
        -------
        Circuit
            A Circuit with stage rounds.
        """
        layers = []
        if self._num_rounds > 1:
            detectors = get_between_round_detectors_and_coordinate_shifts(
                self._detector_coordinates,
            )
            layers.append(
                Circuit(
                    self._construct_syndrome_extraction_layers()
                    + detectors,
                    iterations=self._num_rounds - 1,
                )
            )

        if len(self._final_round_resets) > 0:
            layers.append(GateLayer(self._final_round_resets))

        return Circuit(layers)

    def __eq__(self, other: object) -> bool:
        """
        Calculate whether other is equal to self.

        Parameters
        ----------
        other : object
            Other object for which to calculate equality with self.

        Returns
        -------
        bool
            Whether other and self are equal.
        """
        if isinstance(other, CSSStage):
            return (
                self._stabilisers == other._stabilisers
                and self._num_rounds == other._num_rounds
                and self._first_round_measurements == other._first_round_measurements
                and self._final_round_resets == other._final_round_resets
                and self._observable_definitions == other._observable_definitions
                and self._first_round_gates == other._first_round_gates
            )
        return False



# (c) Copyright Riverlane 2020-2025.
"""Description of ``deltakit.explorer.analysis`` namespace here."""

from deltakit_explorer.analysis._analysis import (
    calculate_lep_and_lep_stddev,
    get_exp_fit,
    get_lambda_fit,
)
from deltakit_explorer.analysis._lambda import calculate_lambda_and_lambda_stddev
from deltakit_explorer.analysis._leppr import (
    LogicalErrorProbabilityPerRoundResults,
    compute_logical_error_per_round,
    simulate_different_round_numbers_for_lep_per_round_estimation,
)
from deltakit_explorer.analysis._quops import (
    predict_distance_for_quops,
    predict_quops_at_distance,
)

from . import error_budget

# List only public members in `__all__`.
__all__ = [s for s in dir() if not s.startswith("_")]



from collections.abc import Mapping
from typing import Protocol

from deltakit_circuit import Circuit
from deltakit_circuit.gates import PauliBasis

from deltakit_explorer.codes import RotatedPlanarCode, css_code_memory_circuit


class MemoryGenerator(Protocol):
    def __call__(self, distance: int, num_rounds: int) -> Circuit: ...


class PreComputedMemoryGenerator(MemoryGenerator):
    def __init__(self, circuits: Mapping[int, Mapping[int, Circuit]]) -> None:
        """A memory generator that used pre-computed circuits.

        Args:
            circuits: a mapping from distance values to another mapping that maps num_rounds values
                to actual quantum circuits. Will be used as ``circuits[distance][num_rounds]``.
        """
        super().__init__()
        self._circuits = circuits

    def __call__(self, distance: int, num_rounds: int) -> Circuit:
        if distance not in self._circuits:
            msg = f"No circuit provided for {distance=}."
            raise RuntimeError(msg)
        circuits_of_distance = self._circuits[distance]
        if num_rounds not in circuits_of_distance:
            msg = f"No circuit provided for {num_rounds=}."
            raise RuntimeError(msg)
        return circuits_of_distance[num_rounds]


def get_rotated_surface_code_memory_circuit(distance: int, num_rounds: int) -> Circuit:
    """Returns a rotated surface code Z memory experiment."""
    return css_code_memory_circuit(
        RotatedPlanarCode(distance, distance), num_rounds, PauliBasis.Z
    )



# (c) Copyright Riverlane 2020-2025.
"""Authentication-related functions and variables."""
from __future__ import annotations

import os

from deltakit_explorer._utils import _utils as utils

TOKEN_VARIABLE = "DELTAKIT_TOKEN"  # nosec B105
TLS_DISABLE_CHECK_VARIABLE = "DELTAKIT_DISABLE_TLS_CHECK"
# pylint: disable=fixme
# TODO: domain name for QEC portal.
RIVERLANE_PORTAL = "https://deltakit.rivelane.com/dashboard/token"


def get_token() -> str:
    """
    Load the token from drive (if any is persisted) to env.
    If there is no token on the drive, read environment variable.
    Return the token.

    Returns:
        str: an auth token.

    Raises:
        RuntimeError: if token is not present at all.
    """
    utils.load_environment_variables_from_drive()
    token = os.environ.get(TOKEN_VARIABLE)
    if token is None:
        file = utils.get_config_file_path()
        msg = (
            f"Token could not be found neither in ({file}) "
            f"nor environment variable ({TOKEN_VARIABLE}). "
            f"Please obtain your token at {RIVERLANE_PORTAL} "
            "and use `Client.set_token` function to "
            "register it."
        )
        raise RuntimeError(
            msg
        )
    return token


def set_token(token: str):
    """
    Set a new token as env variable.

    Args:
        token (str): a string auth token.
    """
    update_dict = {TOKEN_VARIABLE: token}
    utils.set_variables(update_dict, True)
    utils.merge_variables(update_dict, utils.get_config_file_path())


def https_verification_disabled() -> bool:
    """
    In the debug mode you may switch off TLS certificate validation.

    Returns:
        bool: True, if config file DELTAKIT_DISABLE_TLS_CHECK=1|yes|true
    """
    val = os.environ.get(TLS_DISABLE_CHECK_VARIABLE, "0").lower()
    return val in ["1", "yes", "true"]


def set_https_verification(enabled: bool):
    """
    In the debug mode you may switch off TLS certificate validation.

    Args:
        enabled (bool): if set to False, disables HTTPS check.
    """
    val = str(not enabled).lower()
    os.environ[TLS_DISABLE_CHECK_VARIABLE] = val
    utils.set_variables({TLS_DISABLE_CHECK_VARIABLE: val})



from collections.abc import Callable, Sequence

import numpy as np
import numpy.typing as npt
from deltakit_circuit._circuit import Circuit
from deltakit_decode._mwpm_decoder import PyMatchingDecoder
from deltakit_decode.analysis._matching_decoder_managers import StimDecoderManager

from deltakit_explorer.analysis import (
    simulate_different_round_numbers_for_lep_per_round_estimation,
)
from deltakit_explorer.analysis._analysis import calculate_lep_and_lep_stddev
from deltakit_explorer.analysis.error_budget._memory import (
    MemoryGenerator,
    get_rotated_surface_code_memory_circuit,
)


def compute_ideal_rounds_for_noise_model_and_distance(
    noise_model: Callable[[Circuit, npt.NDArray[np.floating]], Circuit],
    noise_parameters: npt.NDArray[np.floating] | Sequence[float],
    distance: int,
    max_shots: int,
    batch_size: int,
    initial_round_number: int = 2,
    min_fails: int = 100,
    target_stddev: float = 1e-4,
    max_round_number: int = 1024,
    next_round_number_func: Callable[[int], int] = lambda x: 4 * x,
    memory_generator: MemoryGenerator = get_rotated_surface_code_memory_circuit,
) -> list[int]:
    """Compute the ideal rounds to use to estimate the LEP per round.

    This function tries to efficiently find the ideal values for the number of rounds to
    use in order to estimate the logical error probability per round. It essentially
    wraps :func:`simulate_different_round_numbers_for_lep_per_round_estimation`,
    using a memory experiment as returned by the provided ``memory_generator``.

    Args:
        noise_model (Callable[[Circuit, npt.NDArray[np.floating]], Circuit]): a callable
            adding noise to the provided circuit, according to the parameters provided.
        noise_parameters (npt.NDArray[numpy.floating] | Sequence[float]): valid
            parameters to forward to ``noise_model`` representing the point at which the
            ideal number of rounds should be computed.
        distance (int): code distance for which we want to have the ideal numbers of
            rounds to estimate the logical error probability per round.
        max_shots (int): maximum number of shots performed by the simulations in this
            function. Simulations might perform less shots due to other conditions being
            met, for example a low-enough standard deviation according to
            ``target_stddev``.
        batch_size (int): number of shots to perform per batch. Early-stopping conditions
            are checked after each chunks of ``batch_size`` shots.
        initial_round_number (int): number of rounds to start the exploration with.
            Should be strictly positive. Should likely not be ``1`` because data for
            ``1`` round is often an outlier. Only set this to ``1`` if you understand
            what you are doing and really want it.
        min_fails (int): minimum number of fails that should be observed to be able to
            early-return before ``max_shots`` shots have been performed.
        target_stddev (float): if the standard deviation of logical error probability
            estimation is below that threshold, simulation might early exit.
        max_round_number (int): maximum number of rounds that should be tested.
        next_round_number_func (Callable[[int], int]): an arbitrary callable that should
            return the next number of rounds to simulate from the previous one. It
            effectively describes the rounds that will be tested until the number of
            rounds is high enough to output a logical error probability over ``0.2`` or
            to exceed ``max_round_number``. Default to a geometric progression with a
            multiplicative factor of ``4``, but can be anything provided that it is
            strictly increasing.
        memory_generator (MemoryGenerator): a callable returning a memory experiment
            from provided distance and number of rounds.

    Returns:
        a list containing number of rounds resulting from successive applications of
        ``next_round_number_func`` starting with ``initial_round_number`` (let
        ``f = next_round_number_func`` and ``x = initial_round_number``, the returned
        list will be ``[x, f(x), f(f(x)), ..., y])``) such that ``y``, the last value of
        that list, is below ``max_round_number`` and ideally lead to a logical error
        probability for ``y`` rounds that is above the ``0.2`` heuristic threshold.
    """

    # Ensure that noise_parameters is a numpy array.
    noise_parameters = np.asarray(noise_parameters)

    def generate_surface_code_memory_and_run(
        num_rounds: int,
    ) -> tuple[int, int]:
        circuit = memory_generator(distance, num_rounds)
        noisy_circuit = noise_model(circuit, noise_parameters)
        decoder, decoder_circuit = PyMatchingDecoder.construct_decoder_and_stim_circuit(
            noisy_circuit
        )
        decoder_manager = StimDecoderManager(decoder_circuit, decoder)

        nshots, nfails = decoder_manager.run_batch_shots(batch_size)
        _, stddev = calculate_lep_and_lep_stddev(nfails, nshots)
        while (stddev > target_stddev or nfails < min_fails) and nshots < max_shots:
            nshots_to_perform = min(batch_size, max_shots - nshots)
            ns, nf = decoder_manager.run_batch_shots(nshots_to_perform)
            nshots += ns
            nfails += nf
            _, stddev = calculate_lep_and_lep_stddev(nfails, nshots)
        return nfails, nshots

    nrounds, *_ = simulate_different_round_numbers_for_lep_per_round_estimation(
        simulator=generate_surface_code_memory_and_run,
        heuristic_logical_error_lower_bound=0.2,
        next_round_number_func=next_round_number_func,
        initial_round_number=initial_round_number,
        maximum_round_number=max_round_number,
    )
    return nrounds.tolist()



# (c) Copyright Riverlane 2020-2025.
"""Description of ``deltakit.explorer.visualisation`` namespace here."""

from deltakit_explorer.plotting._lambda import plot_lambda
from deltakit_explorer.plotting._leppr import plot_logical_error_probability_per_round
from deltakit_explorer.plotting._visualisation import (
                                                            correlation_matrix,
                                                            defect_diagram,
                                                            defect_rates,
)

# List only public members in `__all__`.
__all__ = [s for s in dir() if not s.startswith("_")]



# (c) Copyright Riverlane 2020-2025.
"""
This module consists of functions enabling circuit compilation
using the stim Tableau.
"""
# pylint: disable=too-many-lines,too-many-nested-blocks,too-many-branches,too-many-statements
from collections import defaultdict
from collections.abc import Sequence
from dataclasses import dataclass
from functools import reduce
from operator import mul

import numpy as np
import stim
from deltakit_circuit import Circuit, GateLayer, Layer, Qubit
from deltakit_circuit.gates import (
    CX,
    CXSWAP,
    CY,
    CZ,
    CZSWAP,
    ISWAP,
    ISWAP_DAG,
    MEASUREMENT_GATE_MAPPING,
    MPP,
    MRX,
    MRY,
    MRZ,
    MX,
    MY,
    MZ,
    ONE_QUBIT_GATE_MAPPING,
    RESET_GATE_MAPPING,
    RX,
    RY,
    RZ,
    S_DAG,
    SQRT_XX,
    SQRT_XX_DAG,
    SQRT_YY,
    SQRT_YY_DAG,
    SQRT_ZZ,
    SQRT_ZZ_DAG,
    TWO_QUBIT_GATE_MAPPING,
    XCX,
    XCY,
    XCZ,
    YCX,
    YCY,
    YCZ,
    Gate,
    H,
    I,
    S,
)

from deltakit_explorer.qpu._native_gate_set import (
    NativeGateSetAndTimes,
    OneQubitCliffordGate,
    TwoOperandGate,
)

# key: tuple of the form (gate's layer index, gate's qubit, gate's stim_string representation)
# values: dictionary containing preceding and succeeding unitary block indices:
#   {"preceding": 0, "succeeding": 1}
SpecialGateDict = dict[tuple[int, Qubit, str], dict[str, int]]

# key: tableau in str form: ("+X", "+Z")
# values: tuple of gates that give the tableau, as strs: ("X", "X")
TableauDict = dict[tuple[str, ...], tuple[str, ...]]

# key: tableau, in str form: ("+X", "+Z")
# values: list of possible unitary gates with that tableau: [[I], [X, X], [Z, Z], ...]
EquivalentTableauDict = dict[tuple[str, str], list[list[OneQubitCliffordGate]]]

# A single entry in the CompilationData.two_qubit_gates dictionary. The first
# item in the tuple describes the gate layer index, qubit, and gate.stim_string
# for the given gate. The second entry is the unitary block dictionary, with keys
# being "preceding" or "succeeding" and values being the unitary block index.
TwoQubitGateDictEntry = tuple[tuple[int, Qubit, str], dict[str, int]]

# Dictionary to look up compilation of a two qubit in terms of additional unitary
# gates required. The key can be either a gate currently being compiled or the native
# gate being compiled to. The values are the unitary blocks needed to do the compilation.
# The order is ub1,ub2,ub3,ub4, corresponding to before and after qubit1 and before and
# after qubit 2 respectively.
TwoQubitGateCompilationLookupDict = dict[
    TwoOperandGate,
    tuple[
        list[OneQubitCliffordGate],
        list[OneQubitCliffordGate],
        list[OneQubitCliffordGate],
        list[OneQubitCliffordGate],
    ],
]

MEAS_COMPILATION_LOOKUP_DICT: dict[
    MX | MY | MZ | MRX | MRY | MRZ,
    dict[
        MX | MY | MZ | MRX | MRY | MRZ,
        tuple[list[OneQubitCliffordGate], list[OneQubitCliffordGate]],
    ],
] = {
    MX: {MX: ([], []), MY: ([S], [S_DAG]), MZ: ([H], [H])},
    MY: {MX: ([S_DAG], [S]), MY: ([], []), MZ: ([S_DAG, H], [H, S])},
    MZ: {MX: ([H], [H]), MY: ([H, S], [S_DAG, H]), MZ: ([], [])},
}
# key is the gate you're compiling from, value is the gate
# you're compiling to
RESET_COMPILATION_LOOKUP_DICT: dict[
    RX | RY | RZ,
    dict[
        RX | RY | RZ,
        tuple[list[OneQubitCliffordGate], list[OneQubitCliffordGate]],
    ],
] = {
    RX: {
        RX: ([], MEAS_COMPILATION_LOOKUP_DICT[MX][MX][1]),
        RY: ([], MEAS_COMPILATION_LOOKUP_DICT[MX][MY][1]),
        RZ: ([], MEAS_COMPILATION_LOOKUP_DICT[MX][MZ][1]),
    },
    RY: {
        RX: ([], MEAS_COMPILATION_LOOKUP_DICT[MY][MX][1]),
        RY: ([], MEAS_COMPILATION_LOOKUP_DICT[MY][MY][1]),
        RZ: ([], MEAS_COMPILATION_LOOKUP_DICT[MY][MZ][1]),
    },
    RZ: {
        RX: ([], MEAS_COMPILATION_LOOKUP_DICT[MZ][MX][1]),
        RY: ([], MEAS_COMPILATION_LOOKUP_DICT[MZ][MY][1]),
        RZ: ([], MEAS_COMPILATION_LOOKUP_DICT[MZ][MZ][1]),
    },
}
MEAS_COMPILATION_LOOKUP_DICT[MRX] = {
    MRX: MEAS_COMPILATION_LOOKUP_DICT[MX][MX],
    MRY: MEAS_COMPILATION_LOOKUP_DICT[MX][MY],
    MRZ: MEAS_COMPILATION_LOOKUP_DICT[MX][MZ],
}
MEAS_COMPILATION_LOOKUP_DICT[MRY] = {
    MRX: MEAS_COMPILATION_LOOKUP_DICT[MY][MX],
    MRY: MEAS_COMPILATION_LOOKUP_DICT[MY][MY],
    MRZ: MEAS_COMPILATION_LOOKUP_DICT[MY][MZ],
}
MEAS_COMPILATION_LOOKUP_DICT[MRZ] = {
    MRX: MEAS_COMPILATION_LOOKUP_DICT[MZ][MX],
    MRY: MEAS_COMPILATION_LOOKUP_DICT[MZ][MY],
    MRZ: MEAS_COMPILATION_LOOKUP_DICT[MZ][MZ],
}
# Dictionary to go from the current (CPSWAP-like) gate to CZSWAP. The key
# is the current gate being compiled, and value is a tuple containing
# unitary blocks to do the compilation. The order is ub1,ub2,ub3,ub4,
# corresponding to before and after qubit1 and before and after qubit 2
# respectively. Therefore, it provides the following compilation (in circuit order):
# Gate(a, b) = ub1(a) ub3(b) CZSWAP(a, b) ub2(a) ub4(b)
GATE_TO_CZSWAP_DICT: TwoQubitGateCompilationLookupDict = {
    ISWAP: ([], [S], [], [S]),
    ISWAP_DAG: ([S_DAG], [], [S_DAG], []),
    CXSWAP: ([], [H], [H], []),
    CZSWAP: ([], [], [], []),
}
# Dictionary to go from the CZSWAP to the target gate. The key
# is the target gate, and value is a tuple containing
# unitary blocks to do the compilation. The order is ub1,ub2,ub3,ub4,
# corresponding to before and after qubit1 and before and after qubit 2
# respectively. Therefore, it provides the following compilation (in circuit order):
# CZSWAP(a, b) = ub1(a) ub3(b) Gate(a, b) ub2(a) ub4(b)
CZSWAP_TO_GATE_DICT: TwoQubitGateCompilationLookupDict = {
    ISWAP: ([], [S_DAG], [], [S_DAG]),
    ISWAP_DAG: ([S], [], [S], []),
    CXSWAP: ([], [H], [H], []),
    CZSWAP: ([], [], [], []),
}
# Dictionary to go from the current (CP-like) gate to CZ. The key
# is the current gate being compiled, and value is a tuple containing
# unitary blocks to do the compilation. The order is ub1,ub2,ub3,ub4,
# corresponding to before and after qubit1 and before and after qubit 2
# respectively. Therefore, it provides the following compilation (in circuit order):
# Gate(a, b) = ub1(a) ub3(b) CZ(a, b) ub2(a) ub4(b)
GATE_TO_CZ_DICT: TwoQubitGateCompilationLookupDict = {
    # qubit0_bef, qubit0_aft, qubit1_bef, qubit1_aft
    CX: ([], [], [H], [H]),
    CY: ([], [], [S_DAG, H], [H, S]),
    CZ: ([], [], [], []),
    SQRT_XX: ([H], [S, H], [H], [S, H]),
    SQRT_XX_DAG: ([H], [S_DAG, H], [H], [S_DAG, H]),
    SQRT_YY: ([S_DAG, H], [S, H, S], [S_DAG, H], [S, H, S]),
    SQRT_YY_DAG: ([S_DAG, H], [S, H, S], [S, H], [S, H, S_DAG]),
    SQRT_ZZ: ([], [S], [], [S]),
    SQRT_ZZ_DAG: ([], [S_DAG], [], [S_DAG]),
    XCX: ([H], [H], [H], [H]),
    XCY: ([H], [H], [S_DAG, H], [H, S]),
    XCZ: ([H], [H], [], []),
    YCX: ([S_DAG, H], [H, S], [H], [H]),
    YCY: ([S_DAG, H], [H, S], [S_DAG, H], [H, S]),
    YCZ: ([S_DAG, H], [H, S], [], []),
}
# Dictionary to go from the CZ to the target gate. The key
# is the target gate, and value is a tuple containing
# unitary blocks to do the compilation. The order is ub1,ub2,ub3,ub4,
# corresponding to before and after qubit1 and before and after qubit 2
# respectively. Therefore, it provides the following compilation (in circuit order):
# CZ(a, b) = ub1(a) ub3(b) Gate(a, b) ub2(a) ub4(b)
CZ_TO_GATE_DICT: TwoQubitGateCompilationLookupDict = {
    CX: ([], [], [H], [H]),
    CY: ([], [], [H, S], [S_DAG, H]),
    CZ: ([], [], [], []),
    SQRT_XX: ([H], [H, S_DAG], [H], [H, S_DAG]),
    SQRT_XX_DAG: (
        [H],
        [H, S],
        [H],
        [H, S],
    ),
    SQRT_YY: (
        [H, S],
        [S_DAG, H, S_DAG],
        [H, S],
        [S_DAG, H, S_DAG],
    ),
    SQRT_YY_DAG: (
        [H, S],
        [S_DAG, H, S_DAG],
        [H, S_DAG],
        [S, H, S_DAG],
    ),
    SQRT_ZZ: ([], [S_DAG], [], [S_DAG]),
    SQRT_ZZ_DAG: (
        [],
        [S],
        [],
        [S],
    ),
    XCX: ([H], [H], [H], [H]),
    XCY: ([H], [H], [H, S], [S_DAG, H]),
    XCZ: ([H], [H], [], []),
    YCX: ([H, S], [S_DAG, H], [H], [H]),
    YCY: (
        [H, S],
        [S_DAG, H],
        [H, S],
        [S_DAG, H],
    ),
    YCZ: ([H, S], [S_DAG, H], [], []),
}


def _get_tableau_from_sequence_of_gates(
    unitary_block: Sequence[OneQubitCliffordGate],
) -> stim.Tableau:
    return reduce(
        mul,
        (
            stim.Tableau.from_named_gate(gate.stim_string)
            for gate in unitary_block[::-1]
            # ensuring circuit order, since lestim assumes matrix order for mul
        ),
        stim.Tableau.from_named_gate("I"),  # default in case unitary_block empty
    )


def _get_tableau_as_key(
    tableau: stim.Tableau, up_to_paulis: bool = False
) -> tuple[str, str]:
    """
    Given a stim.Tableau, corresponding to a one-qubit
    Clifford unitary, return a tuple of the string output
    of the Tableau to create a hashable key. Uses the
    stim.Tableau.x_output, stim.Tableau.z_output, in that order
    to produce the output.
    Returns the output with or without sign information, as per
    up_to_paulis.
    E.g, ("+X", "+Z") if up_to_paulis=False, or ("X", "Z") otherwise.

    Parameters
    ----------
    tableau : stim.Tableau
        The tableau to be turned into a hashable key.
    up_to_paulis : bool, optional
        Specify whether signs should be included in the key.

    Returns
    -------
    Tuple[str, str]
        A tuple of the string values of the tableau output.
        E.g, ("+X", "+Z"), or ("X", "Z") if up_to_paulis=True.
    """
    return (
        str(tableau.x_output(0))[int(up_to_paulis) :],
        str(tableau.z_output(0))[int(up_to_paulis) :],
    )


def _get_tableau_from_sequence_of_1q_gates(gates: Sequence[str]) -> stim.Tableau:
    if len(gates) == 0:
        return stim.Tableau.from_named_gate("I")
    try:
        return reduce(
            mul,
            (
                stim.Tableau.from_named_gate(gate)
                for gate in gates[::-1]
                # ensuring circuit order, since lestim assumes matrix order for mul
            ),
        )
    except ValueError as ve:
        msg = "'gates' must be composed of only single qubit gates"
        raise ValueError(msg) from ve


def _get_compilation_dict(
    native_gates: NativeGateSetAndTimes,
    max_length: int | None = None,
    up_to_paulis: bool = False,
) -> tuple[TableauDict, EquivalentTableauDict]:
    """
    Create a compilation dictionary that specifies, for a given set of single
    qubit native gates, the lowest length operator-product per tableau. To achieve this,
    starting from the identity, multiply in all the native single-qubit gates to create
    length one operators and add the unique tableaus to a dictionary, with tableau as
    key and corresponding operator as value. Then progressively multiply in the single-
    qubit native gates creating length 2, 3 etc. operators, keeping only the minimum-length
    ones for a unique set of tableaus, up to the maximum of 24, or until no new tableaus
    are found for a given operator length.

    Parameters
    ----------
    native_gates : NativeGateSetAndTimes
        Native gate set. If `one_qubit_gates` is empty, return
        a dictionary only containing the identity.
    max_length : Optional[int], optional
        Maximum length of operators to be considered, used to put an upper
        limit on which combinations of the single-qubit native gates should
        be created.
        By default, None, which will let the program run until it either
        finds all tableaus or determines there are none left to find.
    up_to_paulis : bool, optional
        If True, then when creating the Tableau dictionary, only consider
        Tableaus that are equivalent up to Pauli terms. This means not
        considering the sign of the conjugation result. The maximum number
        of unique Tableaus then becomes 6.
        By default, False.

    Returns
    -------
    Tuple[TableauDict, EquivalentTableauDict]
        A tuple containing the compilation dictionary, and the equivalent
        tableau dictionary.

        Compilation dictionary: TableauDict
            A dictionary with Tableaus as keys and operators as values.
            E.g, `{("+X", "-Z"): ("X",), ("+X", "+Z"): ()}`
            The tableau is created from stim.Tableau.x_output and
            stim.Tableau.z_output, in that order.
            The value is the empty tuple if the tableau key is equivalent to the identity.
        Equivalent tableau dictionary: EquivalentTableauDict
            A dictionary with Tableaus as keys and lists of lists of OneQubitCliffordGate
            as values, where the list expresses the gates in circuit order.
            E.g, `{("+Z", "-X"): [[H, Z], [X, H]], ...}`
    """

    if up_to_paulis:
        max_unique_tableau_count = 6
        identity_tableau = ("X", "Z")
        one_q_gates_as_stim_string = {
            str(g.stim_string)
            for g in native_gates.one_qubit_gates
            if str(g.stim_string) not in ["X", "Y", "Z"]
        }
    else:
        max_unique_tableau_count = 24
        identity_tableau = ("+X", "+Z")
        one_q_gates_as_stim_string = {
            str(g.stim_string) for g in native_gates.one_qubit_gates
        }

    # Include the identity tableau, as it is always there by default
    min_weight_tableau_dict: TableauDict = {identity_tableau: ()}
    equiv_tableau_dict: defaultdict[
        tuple[str, str], list[list[OneQubitCliffordGate]]
    ] = defaultdict(list)
    current_weight = 1
    while max_length is None or current_weight <= max_length:
        # keep track of whether we've inserted any new tableaus. if we haven't,
        # we can stop the computation there
        new_tableau_found = False
        # if True, return dicts at end of current gate product iteration weight.
        # e.g, if set True on weight 3 products, will iterate through all the remaining
        # weight 3 products before returning. this is to catch equivalent tableaus remaining
        finish_at_end_of_product_loop = False

        # need to make a copy since we will change the dictionary during the loop
        copy_of_dict_to_iterate_over = min_weight_tableau_dict.copy()
        for current_gates in copy_of_dict_to_iterate_over.values():
            if len(current_gates) != current_weight - 1:
                continue
            for native_gate in one_q_gates_as_stim_string:
                gate_product = (*current_gates, native_gate)
                product_tableau = _get_tableau_from_sequence_of_1q_gates(gate_product)
                tableau_key = _get_tableau_as_key(product_tableau, up_to_paulis)

                # if new tableau found, add it, it will be minimum weight
                if tableau_key not in min_weight_tableau_dict:
                    min_weight_tableau_dict[tableau_key] = gate_product
                    equiv_tableau_dict[tableau_key].append(
                        [ONE_QUBIT_GATE_MAPPING[g] for g in gate_product]
                    )
                    new_tableau_found = True
                    if len(min_weight_tableau_dict) == max_unique_tableau_count:
                        # we've found all possible tableaus
                        # dont return too soon, as we want to find equivalent tableaus
                        finish_at_end_of_product_loop = True

                # if tableau already in dict, consider it for equivalent tableau dict
                # only add equivalent gate expressions of minimum length
                elif len(gate_product) <= len(min_weight_tableau_dict[tableau_key]):
                    equiv_tableau_dict[tableau_key].append(
                        [ONE_QUBIT_GATE_MAPPING[g] for g in gate_product]
                    )

        if not new_tableau_found or finish_at_end_of_product_loop:
            # since we didn't find any new tableaus, we can stop
            return min_weight_tableau_dict, equiv_tableau_dict
        current_weight += 1
    return min_weight_tableau_dict, equiv_tableau_dict


def _get_tableau_z_image_as_string(tableau: stim.Tableau, up_to_paulis: bool):
    return str(tableau.z_output(0))[int(up_to_paulis) :]


def _get_tableau_x_image_as_string(tableau: stim.Tableau, up_to_paulis: bool):
    return str(tableau.x_output(0))[int(up_to_paulis) :]


def _get_tableau_y_image_as_string(tableau: stim.Tableau, up_to_paulis: bool):
    return str(tableau.y_output(0))[int(up_to_paulis) :]


def _get_compilation_with_projectors_before_unitaries(
    compilation_dict: TableauDict,
    unitary_block: Sequence[OneQubitCliffordGate],
    projector_before_unitaries: MZ | RZ | MX | RX | MY | RY,
    up_to_paulis: bool = False,
) -> tuple[str, ...]:
    """
    Given a string/block of unitary gates, and specifying which projector
    this unitary block is preceded by (reset or measurement), use the Tableau
    to work out if this set of projector + unitaries can be reduced to a
    shorter set of gates. This requires comparing Tableaus that are identical in the
    conjugation corresponding to the basis of the projector. E.g, if two Tableaus
    conjugate +Z to +X, then the Tableau with the shortest gate representation may
    be chosen to follow a reset or measurement in the Z basis. Likewise for the other
    bases. If the given set of unitaries is empty, the Identity will be returned.
    If the compilation dictionary contains no tableau that matches that of the
    `unitary_block`, then a KeyError is thrown, as this may cause a return value in
    terms of non-native gates.
    If the compilation is requested in the Y basis, e.g, MY or RY is provided, the
    Y tableau of the dictionary components will be computed from their X and Z parts.

    Parameters
    ----------
    compilation_dict : TableauDict
        Tableau dict calculated for the native gate set corresponding to
        the circuit being compiled. Values (gates) follow circuit order.
    unitary_block : Sequence[OneQubitCliffordGate]
        Unitary gates to try to reduce by comparing Tableaus, specified
        in circuit order.
    projector_before_unitaries : Union[MZ, RZ, MX, RX, MY, RY]
        Specify which projector is immediately before
        the unitary_block in the circuit.
    up_to_paulis : bool, Optional
        If True, the tableau computed for the unitary block
        will be computed only up to Paulis, meaning the sign information
        is not considered in comparisons with `compilation_dict`.
        The provided `compilation_dict` should match with the choice
        of up_to_paulis.
        False by default.

    Returns
    -------
    Tuple[str, ...]
        Shortest length gate series corresponding to identical action under
        Tableau for the relevant basis of projector_before_unitaries.
    """
    # first, get the Tableau for the unitary_block
    if len(unitary_block) == 0:
        unitary_block_tableau: stim.Tableau = stim.Tableau.from_named_gate("I")
        shortest_gates: tuple = ()
    else:
        unitary_block_tableau = _get_tableau_from_sequence_of_1q_gates(
            [gate.stim_string for gate in unitary_block]
        )
        shortest_gates = tuple(str(g.stim_string) for g in unitary_block)

    # use appropriate function to get tableau image depending on basis
    if projector_before_unitaries in [MZ, RZ]:
        get_tableau_image_as_string = _get_tableau_z_image_as_string
        tableau_index = 1
    elif projector_before_unitaries in [MX, RX]:
        get_tableau_image_as_string = _get_tableau_x_image_as_string
        tableau_index = 0
    elif projector_before_unitaries in [MY, RY]:
        get_tableau_image_as_string = _get_tableau_y_image_as_string
        # update tableau to include Y term at index 2
        tableau_index = 2
        compilation_dict_with_y = compilation_dict.copy()
        for tableau, gates in list(compilation_dict.items()):
            compilation_dict_with_y[
                (
                    *tableau,
                    str(
                        stim.Tableau.from_conjugated_generators(
                            xs=[stim.PauliString(tableau[0])],
                            zs=[stim.PauliString(tableau[1])],
                        ).y_output(0)
                    )[
                        1 if up_to_paulis else 0 :
                    ],  # remove sign if not needed
                )
            ] = gates
            del compilation_dict_with_y[tableau]
        compilation_dict = compilation_dict_with_y
    else:
        msg = f"{projector_before_unitaries.stim_string} is not a recognised projector"
        raise NotImplementedError(msg)

    # for the basis of the reset or measurement before a unitary block, we can just consider
    # the part of the Tableau relevant to the basis.
    unitary_block_tableau_in_comp_dict = False
    for tableau, gates in compilation_dict.items():
        if tableau[tableau_index] == get_tableau_image_as_string(
            unitary_block_tableau, up_to_paulis
        ):
            unitary_block_tableau_in_comp_dict = True
            if len(gates) < len(shortest_gates):
                shortest_gates = gates

    if not unitary_block_tableau_in_comp_dict:
        msg = (
            "unitary_block's tableau is not in the compilation_dictionary."
            " This means the output of this function may include gates not"
            " in the native gate set. Try compiling unitary_block to the"
            " native gate set first."
        )
        raise KeyError(msg)

    return shortest_gates


def _get_compilation_with_measurement_after_unitaries(
    compilation_dict: TableauDict,
    unitary_block: Sequence[OneQubitCliffordGate],
    measurement_after_unitaries: MZ | RZ | MX | RX | MY | RY,
    up_to_paulis: bool = False,
) -> tuple[str, ...]:
    """
    In the case of a measurement following a block of unitary operators,
    we may simply consider the Tableau's that treat the measurement basis
    Pauli in an equivalent way. To this end, we consider the Tableau of
    the unitary block, and see under which conjugation the measurement
    basis is maintained. Then, we search the compilation dictionary for
    equivalent Tableaus that offer the same conjugation. This is achieved
    by calling `_get_compilation_with_projectors` with `projector_before_unitaries`
    chosen to be the conjugation that gives the unchanged measurement basis.

    Parameters
    ----------
    compilation_dict : TableauDict
        Tableau dict calculated for the native gate set corresponding to
        the circuit being compiled.
    unitary_block : Collection[OneQubitCliffordGate]
        Unitary gates to try to reduce by comparing Tableaus.
    measurement_after_unitaries : Union[MZ, MX, MY]
        Specify which measurement is immediately after
        the unitary_block in the circuit.
    up_to_paulis : bool, Optional
        If True, the tableau computed for the unitary block
        will be computed only up to Paulis, meaning the sign information
        is not considered in comparisons with `compilation_dict`.
        The provided `compilation_dict` should match with the choice
        of up_to_paulis.
        False by default.

    Returns
    -------
    Tuple[str, ...]
        Shortest length gate series corresponding to identical action under
        Tableau for the relevant basis of measurement_after_unitaries.
    """
    unitary_block_tableau = _get_tableau_from_sequence_of_gates(unitary_block)
    if (
        str(unitary_block_tableau.x_output(0))[1:]
        == measurement_after_unitaries.basis.value
    ):
        new_basis = RX
    elif (
        str(unitary_block_tableau.z_output(0))[1:]
        == measurement_after_unitaries.basis.value
    ):
        new_basis = RZ
    else:
        new_basis = RY

    return _get_compilation_with_projectors_before_unitaries(
        compilation_dict, unitary_block, new_basis, up_to_paulis
    )


def _is_identity_like(tableau: stim.Tableau) -> bool:
    """
    Function to test whether a tableau is identity-like. This is achieved via
    comparing to the identity Tableau in the numpy form of the tableau.
    If a tableau is identity like, the X-to-X and Z-to-Z components will be
    the same as the identity matrix, and the X-to-Z and Z-to-X components will
    be equal to the all-zero matrix. If this is not the case, we deduce that
    the tableau is not identity-like.

    Parameters
    ----------
    tableau : stim.Tableau
        Tableau to compare to identity.

    Returns
    -------
    bool
        True if the tableau is identity-like, false otherwise.
    """
    x2x, x2z, z2x, z2z, _, _ = tableau.to_numpy()
    for identity_component in [x2x, z2z, x2z, z2x]:
        # check if any off-diagonal elements are non-zero
        if (identity_component[~np.eye(len(identity_component), dtype=np.bool_)]).any():
            return False
    return True


def _get_tableau_key_from_sequence_of_gates(
    gates: Sequence[OneQubitCliffordGate], up_to_paulis: bool = False
) -> tuple[str, str]:
    return _get_tableau_as_key(
        _get_tableau_from_sequence_of_1q_gates([gate.stim_string for gate in gates]),
        up_to_paulis,
    )


def _get_single_qubits_tableau_key_from_two_qubit_tableau(
    two_qubit_tableau: stim.Tableau, qubit_index: int, up_to_paulis: bool = False
) -> tuple[str, str]:
    """
    Given a tableau for a two-qubit gate, get the part of the tableau that corresponds
    to just one of the qubits, specified by qubit_index.
    May also specify up_to_paulis to get the key without signs.

    Parameters
    ----------
    two_qubit_tableau : stim.Tableau
        Tableau for the two qubit gate.
    qubit_index : int
        Index of the single qubit portion of the tableau to return.
    up_to_paulis : bool, optional
        Specify whether signs should be included in the key or not.

    Returns
    -------
    Tuple[str, str]
        Tableau key for the qubit_index part of two_qubit_tableau.
        E.g, ("+Z", "+X") or ("Z", "X") if up_to_paulis=True.
    """
    if two_qubit_tableau.to_unitary_matrix(endian="little").shape != (4, 4):
        msg = "The given tableau does not describe a two qubit gate"
        raise ValueError(msg)
    if qubit_index == 0:
        tableau_key = (
            str(two_qubit_tableau.x_output(qubit_index))[
                :-1  # looks like "+Z_", so remove underscore here
            ][
                int(up_to_paulis) :
            ],  # remove +/- if up_to_paulis True
            str(two_qubit_tableau.z_output(qubit_index))[:-1][int(up_to_paulis) :],
        )
    else:
        tableau_key = (
            str(two_qubit_tableau.x_output(1))[
                0:3:2  # looks like "+_Z", so remove underscore here
            ][
                int(up_to_paulis) :
            ],  # remove +/- if up_to_paulis True
            str(two_qubit_tableau.z_output(1))[0:3:2][int(up_to_paulis) :],
        )
    return tableau_key


def _get_compilation_with_two_qubit_gates(
    two_qubit_gate: TwoOperandGate,
    compilation_dict: TableauDict,
    unitary_block_1: list[OneQubitCliffordGate],
    unitary_block_2: list[OneQubitCliffordGate],
    unitary_block_3: list[OneQubitCliffordGate],
    unitary_block_4: list[OneQubitCliffordGate],
    gate_exchange_dict: EquivalentTableauDict | None = None,
    up_to_paulis: bool = False,
    allow_terms_to_mutate: bool = True,
    allow_terms_to_multiply: bool = True,
) -> tuple[
    list[OneQubitCliffordGate],
    list[OneQubitCliffordGate],
    list[OneQubitCliffordGate],
    list[OneQubitCliffordGate],
]:
    """
    Given the following circuit diagram:
                          ___
    control: ---| U1 |---|   |---| U2 |---
                         |   |
                         | G |
                         |   |
    target:  ---| U3 |---|___|---| U4 |---

    Compile a two-qubit gate G and its surrounding one-qubit unitary blocks
    U1, U2, U3, U4 to reduce the number of unitaries before the gate,
    at the cost of increasing the number of unitaries after the gate,
    which may allow further compilation of the unitaries after the
    gate to reduce the overall unitary gate count.
    U1, U2, U3, U4 are in chronological order.

    Parameters
    ----------
    two_qubit_gate : TwoOperandGate
        The two-qubit gate inbetween the unitary blocks.
    compilation_dict : TableauDict
        The compilation dictionary for this circuit. Used to reduce unitary
        blocks to possibly shorter sequences.
    unitary_block_1 : List[OneQubitCliffordGate]
        The unitary block in the top left of the diagram, before the first qubit.
    unitary_block_2 : List[OneQubitCliffordGate]
        The unitary block in the top right of the diagram, after the first qubit.
    unitary_block_3 : List[OneQubitCliffordGate]
        The unitary block in the bottom left of the diagram, before the second qubit.
    unitary_block_4 : List[OneQubitCliffordGate]
        The unitary block in the bottom right of the diagram, after the second qubit.
    gate_exchange_dict : EquivalentTableauDict, optional
        Dictionary to provide extra possible gate combinations for a given Tableau.
        Used to exchange unitary blocks for equivalent Tableaus to check if there is
        an alternative set of gates that may allow some further simplification.
        By default, empty dictionary.
    up_to_paulis : bool, optional
        If True, will consider tableaus equivalent if they are equivalent up
        to pauli terms. E.g, the sign of the tableau terms are not considered.
    allow_terms_to_mutate : bool, optional
        If True, will allow the case of pulling a single-qubit Pauli term through a
        two-qubit gate such that the resulting (possibly two-qubit) Pauli has a different
        term on either qubit.
        E.g, pulling an X through a CZ from the first qubit mutates into a Z term on the second qubit.
        By default, True.
    allow_terms_to_multiply : bool, optional = True
        If True, will allow the case of pulling a term through a two-qubit gate
        such that a second term on the other qubit is created. E.g, pulling an X through
        a CX from the first qubit mutates into an X on the first qubit and an X on the
        second qubit.
        By default, True.

    Returns
    -------
    Tuple[
        List[OneQubitCliffordGate],
        List[OneQubitCliffordGate],
        List[OneQubitCliffordGate],
        List[OneQubitCliffordGate],
    ]:
        Tuple containing the updated unitary blocks 1-4 in order.
    """
    if gate_exchange_dict is None:
        gate_exchange_dict = {}

    # get tableau and inverse tableau of two_qubit_gate
    two_qubit_tableau = stim.Tableau.from_named_gate(two_qubit_gate.stim_string)  # G
    two_qubit_tableau_inv = two_qubit_tableau.inverse()  # G_adj

    # iterate over these to reduce code duplication.
    # first or second qubit changes which parts we look at in some places
    for unitary_block_whole, is_second_qubit in zip(
        [unitary_block_1, unitary_block_3], [False, True]
    ):
        # track how many gates have been pulled through, to adjust what we look at
        # e.g, if we took 2 gates off the end for compilation, this increases by 2,
        # to stop us looking at those 2 gates again.
        offset = 0

        # we first look at the rightmost gate in the unitary block to see if it can
        # be pulled through. If not, we look at the 2 rightmost gates, etc until
        # we have considered unitary block as a whole.
        #  ...--A1-A2-A3--  |  outer_start_index
        #             ^        = len-1
        #          ^  ^        = len-2
        #       ^  ^  ^        = len-3
        for outer_start_index in range(len(unitary_block_whole) - 1, -1, -1):
            # within each sub-block of gates in unitary_block, see if there
            # is an equivalent gate expression that lets us pull something through
            unitary_block_equiv_gates: list[
                list[OneQubitCliffordGate]
            ] = gate_exchange_dict.get(
                _get_tableau_key_from_sequence_of_gates(
                    unitary_block_whole[
                        outer_start_index : len(unitary_block_whole) - offset
                    ],
                    up_to_paulis,
                ),
                [
                    unitary_block_whole[
                        outer_start_index : len(unitary_block_whole) - offset
                    ]
                ],
            )

            # e.g, for unitary_block in [[SQRT_X, S, SQRT_X], [S, SQRT_X, S], ...]
            for current_unitary_block_equiv in unitary_block_equiv_gates:
                # for each of the equivalent expressions of unitary_block we examine,
                # we should also iterate over it starting from the right, the same as the
                # outer loop, to catch cases where only a part of the equivalent expression
                # can be pulled through
                for inner_start_index in range(
                    len(current_unitary_block_equiv) - 1, -1, -1
                ):
                    # we look at the portion of unitary_block that starts from inner_start_index onwards
                    partial_unitary_block = current_unitary_block_equiv[
                        inner_start_index:
                    ]

                    # get tableau for unitary block A, and its two qubit version AI or IA,
                    # by adding it to the identity, in the right order dependent on if
                    # unitary_block is the first or second qubit
                    unitary_block_tab = _get_tableau_from_sequence_of_1q_gates(
                        [gate.stim_string for gate in partial_unitary_block]
                    )
                    two_qubit_unitary_block_tab = (
                        stim.Tableau.from_named_gate("I") + unitary_block_tab
                        if is_second_qubit
                        else unitary_block_tab + stim.Tableau.from_named_gate("I")
                    )

                    # get tableau of G_adj A G, for G our two qubit gate
                    # if this is not identity-like, stop
                    conj_a_with_2q_gate_and_inv = (
                        two_qubit_tableau
                        * two_qubit_unitary_block_tab
                        * two_qubit_tableau_inv
                    )

                    # we can only pull things through if the conjugated tableau is identity-like
                    if not _is_identity_like(conj_a_with_2q_gate_and_inv):
                        continue
                    # check for the two optional conditions
                    if not allow_terms_to_mutate:
                        two_qubit_unitary_block_tab_reverse = (
                            unitary_block_tab + stim.Tableau.from_named_gate("I")
                            if is_second_qubit
                            else stim.Tableau.from_named_gate("I") + unitary_block_tab
                        )
                        if (
                            # check that pulling u through G, u-G becomes only
                            # G-u, with u possibly being on either qubit or both, as long as
                            # the term has not changed. That is, a Z pulled through must remain
                            # as a Z, and not change to any other Pauli term.
                            conj_a_with_2q_gate_and_inv
                            not in (
                                two_qubit_unitary_block_tab,
                                unitary_block_tab + unitary_block_tab,
                                two_qubit_unitary_block_tab_reverse,
                            )
                        ):
                            continue
                    if not allow_terms_to_multiply:
                        identity_tab = ("X", "Z") if up_to_paulis else ("+X", "+Z")
                        qubit_to_check_for_created_term = 0 if is_second_qubit else 1
                        if (
                            _get_single_qubits_tableau_key_from_two_qubit_tableau(
                                conj_a_with_2q_gate_and_inv,
                                qubit_to_check_for_created_term,
                                up_to_paulis,
                            )
                            != identity_tab
                        ):
                            continue

                    # from here, our goal is to calculate W2 and W4 where A G U2 U4 = G W2 W4 is compiled
                    # optimally. We know this is possible as the identity-likeness test passed.

                    # multiply this tableau with u2+u4 and then read off the u2 and u4 parts individually
                    u2_tab = _get_tableau_from_sequence_of_1q_gates(
                        [gate.stim_string for gate in unitary_block_2]
                    )
                    u4_tab = _get_tableau_from_sequence_of_1q_gates(
                        [gate.stim_string for gate in unitary_block_4]
                    )
                    # W2 W4
                    ub_and_2q_tab_with_u2u4 = (
                        u2_tab + u4_tab
                    ) * conj_a_with_2q_gate_and_inv  # matrix order

                    # read conjugation for first and second qubits separately
                    # W2
                    first_qubit_tab = (
                        _get_single_qubits_tableau_key_from_two_qubit_tableau(
                            ub_and_2q_tab_with_u2u4, 0, up_to_paulis
                        )
                    )

                    # W4
                    second_qubit_tab = (
                        _get_single_qubits_tableau_key_from_two_qubit_tableau(
                            ub_and_2q_tab_with_u2u4, 1, up_to_paulis
                        )
                    )

                    # lookup this tableau in the compilation dict to find the relevant gate
                    try:
                        # W2
                        unitary_block_2 = [
                            ONE_QUBIT_GATE_MAPPING[
                                gate
                            ]  # we have the gate string, so change to actual gate class
                            for gate in compilation_dict[first_qubit_tab]
                        ]
                        # W4
                        unitary_block_4 = [
                            ONE_QUBIT_GATE_MAPPING[gate]
                            for gate in compilation_dict[second_qubit_tab]
                        ]
                    except KeyError as ke:
                        msg = f"Tableau missing from compilation dictionary: {ke!s}"
                        raise KeyError(msg) from ke

                    # update pre-2q gate unitaries
                    # = original block up to where our outer index is
                    #   + current version of block up to pulled through gates
                    offset += len(current_unitary_block_equiv) - inner_start_index
                    if is_second_qubit:
                        unitary_block_3 = (
                            unitary_block_3[:outer_start_index]
                            + current_unitary_block_equiv[:inner_start_index]
                        )
                    else:
                        unitary_block_1 = (
                            unitary_block_1[:outer_start_index]
                            + current_unitary_block_equiv[:inner_start_index]
                        )

                    # since we've compiled, we should start again from the new unitary blocks
                    return _get_compilation_with_two_qubit_gates(
                        two_qubit_gate,
                        compilation_dict,
                        unitary_block_1,
                        unitary_block_2,
                        unitary_block_3,
                        unitary_block_4,
                        gate_exchange_dict,
                        up_to_paulis,
                        allow_terms_to_mutate,
                        allow_terms_to_multiply,
                    )

    return (unitary_block_1, unitary_block_2, unitary_block_3, unitary_block_4)


@dataclass
class CompilationData:
    """
    Dataclass to represent the data extracted from a `deltakit.circuit.Circuit`,
    to be used in Tableau compilation of that circuit.

    Parameters
    ----------
    unitary_blocks : Dict[int, List[OneQubitCliffordGate]]
        All the one-qubit unitary blocks in the Circuit, indexed by integers.
        A unitary block is a sequence of one-qubit unitary gates preceding or succeeding
        a reset, measurement or two-qubit gate.
    reset_gates : SpecialGateDict
        A dictionary of all the reset gates in the circuit. Keys are tuples
        of the form (GateLayer index, qubit, gate type) with values being dictionaries
        of the form {"preceding": unitary_block_index, "succeeding": unitary_block_index}
    measurement_gates : SpecialGateDict
        A dictionary of all the measurement gates in the circuit. Keys are tuples
        of the form (GateLayer index, qubit, gate type) with values being dictionaries
        of the form {"preceding": unitary_block_index, "succeeding": unitary_block_index}
    two_qubit_gates : SpecialGateDict
        A dictionary of all the two-qubit gates in the circuit. Keys are tuples
        of the form (GateLayer index, qubit, gate type) with values being dictionaries
        of the form {"preceding": unitary_block_index, "succeeding": unitary_block_index}
    non_gatelayer_layers : Dict[int, Layer]
        A dictionary of the layer index of a non-GateLayer layer. Keys are the layer's index,
        the values are the layer itself.
    iterations : int
        Number of iterations of the circuit, to represent repeat blocks.
    num_layers : int
        Number of layers in the circuit
    """

    unitary_blocks: dict[int, list[OneQubitCliffordGate]]
    reset_gates: SpecialGateDict
    measurement_gates: SpecialGateDict
    two_qubit_gates: SpecialGateDict
    non_gatelayer_layers: dict[int, Layer]
    iterations: int = 1
    num_layers: int = 0


def _get_relevant_dict_to_update(
    gate: Gate,
    reset_dict: SpecialGateDict,
    meas_dict: SpecialGateDict,
    two_q_dict: SpecialGateDict,
) -> SpecialGateDict:
    if isinstance(gate, (RX, RY, RZ)):
        return reset_dict
    if isinstance(gate, (MRX, MRY, MRZ, MX, MY, MZ)):
        return meas_dict
    if isinstance(gate, TwoOperandGate):
        return two_q_dict
    msg = f"Gate is not MR, M, R or 2Q: {gate}"
    raise ValueError(msg)


def _extract_structure_from_circuit(
    circuit: Circuit,
) -> CompilationData:
    """
    From the given circuit, extract the data relevant for compiling the circuit using
    the Tableau. This means we need to have collections of one-qubit unitary blocks
    that precede or succeed reset, measurement and two-qubit gates. In order to
    simplify compilation and reduce implementation complexity, each collection of
    reset, measurement and two-qubit gates will point to a shared memory of unitary
    blocks. This way, when one set of unitary blocks is updated after compilation of,
    for example, resets, the result will automatically propagate forward for when the
    same block is reconsidered but as a part of a different gate, say, a measurement
    later on in the circuit. This removes the complexity of having to keep track of
    which gates have compiled at what time, for which gate etc.

    The data structures are as follows:
        unitary blocks: returned as first item in the tuple.
            Stored as a dictionary of int keys and List[OneQubitCliffordGate] values.
            The key is the index of the unitary block that is pointed to by the reset, measurement
            and two-qubit gate dictionaries. The list is the unitary block itself.
        SpecialGateDict
            This data structure stores the relevant data for the `special` gates, those being
            reset, measurement and two-qubit gates. This dictionary has Tuple[int, Qubit] keys,
            and Dict["preceding": int, "succeeding": int] values.
            The key is (GateLayer index of gate, Qubit of gate, Gate). The value is two integers keyed
            by strings `preceding` and `succeeding` that point to a unitary block via that block's
            index.

        For example, if there is the small circuit
        ---RZ(0)--H(0)--MZ(0)---
        Then there are 3 unitary blocks here:
            (1) the empty block preceding the reset,
            (2) the block inbetween the reset and measurement,
            (3) the block succeeding the measurement
        In this case, unitary_blocks' data is
            {0: [], 1: [H(0)], 2: []}
            with indices corresponding to above labelling.
        Then, the reset dict would be:
            {(0, Qubit(0), RZ): {"preceding": 0, "succeeding": 1}}
        and the measurement dict:
            {(0, Qubit(0), MZ): {"preceding": 1, "succeeding": 2}}
        Where you can see that the reset and measurement dicts simply point to the unitary blocks,
        such that if unitary block 1 is compiled at some point, any changes will be propagated to
        later compilations automatically. If it is updated after compiling resets, the changes
        will be present when measurements are considered.

    The algorithm to collect such data structures from the circuit is as follows:
        For each gate layer:
            For each qubit in layer:
                if it is a special gate:
                    (0) all gates encountered on this qubit so far are the unitary block preceding this gate
                    (1) update the relevant dict: {(layer index, qubit, gate): {"preceding": unitary_block_index}}
                    (2) reset (clear) the gates_encountered dict for this qubit
                    (3) mark current special gate as most recent special gate on this qubit
                    current unitary block may also be a "succeeding" for the most recent special gate on this qubit
                    (4) check most recent special gate on this qubit
                    (5) if exists, update as above but with "succeeding" key
                else:
                    add it to gates encountered on this qubit so far
        If not gate layer, append to none-GateLayer struct.

        do a final update using the remaining encountered gates per qubit, as these
        correspond to gates at the end of the circuit and must be stored as "succeeding"
        of all the last encountered special gates on each qubit.

        if a qubit had no resets, measurements or two-qubit gates on it, then we drop it from the circuit.


    Parameters
    ----------
    circuit : Circuit
        Circuit to parse into a data structure such that
        compilation is possible.

    Returns
    -------
    CompilationData
        A dataclass of the data required for compilation, extracted
        from `circuit`.
    """
    # iterate over the layers
    # {(gate_layer_index, qubit_index, gate_stim_string): {preceding: [], succeeding: []}}
    unitary_blocks: dict[int, list[OneQubitCliffordGate]] = {}
    reset_dict: SpecialGateDict = {}
    meas_dict: SpecialGateDict = {}
    two_q_dict: SpecialGateDict = {}
    non_gate_layers: dict[int, Layer] = {}

    # for each special gate, keep track of any unitaries that have been encountered
    # e.g, for a circuit: q0 = --X--Z--S--RZ--, this would be {q0: [X, Z, S]} up until
    # the reset, at which point those unitaries are set as a unitary block preceding
    # the RZ, then cleared.
    current_gates_per_qubit: dict[Qubit, list[OneQubitCliffordGate]] = {
        q: [] for q in circuit.qubits
    }
    # keep track of the most recent special gate encountered on a qubit. initially set
    # to the identity on qubit -1 at layer -1 as a default value.
    # e.g, after encountering an RZ at layer 0 on Qubit(0), this will be {Qubit(0): (RZ(0), 0)}.
    # if we then later encounter an MZ at layer 5 on Qubit(0), this will be {Qubit(0): (MZ(0), 5)}.
    most_recent_special_gate_per_qubit: dict[Qubit, tuple[Gate, int]] = {
        q: (I(-1), -1) for q in circuit.qubits
    }
    unitary_block_index = 0
    for layer_i, layer in enumerate(circuit.layers):
        # if layer is Circuit, add it to non_gate_layers. this allows us to
        # preserve repeat blocks when unpacking this data structure later
        if isinstance(layer, Circuit):
            # since we've encountered a repeat block, add any trailing unitaries
            # as succeeding blocks to the most recent special gates, and clear
            # the current_gates_per_qubit dict so gates don't carry over beyond
            # the repeat block
            for qubit in circuit.qubits:
                unitary_blocks[unitary_block_index] = current_gates_per_qubit[qubit]
                if not isinstance(most_recent_special_gate_per_qubit[qubit][0], I):
                    dict_to_update = _get_relevant_dict_to_update(
                        most_recent_special_gate_per_qubit[qubit][0],
                        reset_dict,
                        meas_dict,
                        two_q_dict,
                    )
                    dict_to_update[
                        (
                            most_recent_special_gate_per_qubit[qubit][1],
                            qubit,
                            most_recent_special_gate_per_qubit[qubit][0].stim_string,
                        )
                    ]["succeeding"] = unitary_block_index
                else:
                    # this qubit had no meas, reset or 2q gates applied to it
                    continue
                current_gates_per_qubit[qubit] = []
                unitary_block_index += 1
            # need to clear most recent gate per qubit too
            most_recent_special_gate_per_qubit = {
                q: (I(-1), -1) for q in circuit.qubits
            }
            non_gate_layers[layer_i] = _extract_structure_from_circuit(layer)
        elif isinstance(layer, GateLayer):
            # traverse GateLayer
            for gate in layer.gates:
                # if gate is just a unitary, add it and move on
                if isinstance(gate, MPP):
                    msg = "MPP gates not yet supported"
                    raise NotImplementedError(msg)
                if not isinstance(
                    gate, (RX, RY, RZ, MRX, MRZ, MRY, MX, MY, MZ, TwoOperandGate)
                ):
                    current_gates_per_qubit[gate.qubit].append(gate)
                    continue
                # from here we know it is a special gate.
                # load current unitary block based on gates encountered on this qubit so far
                for qubit in gate.qubits:
                    unitary_blocks[unitary_block_index] = current_gates_per_qubit[qubit]

                    # check which of the special gates `gate` is, and update
                    dict_to_update = _get_relevant_dict_to_update(
                        gate, reset_dict, meas_dict, two_q_dict
                    )
                    dict_to_update[(layer_i, qubit, gate.stim_string)] = {
                        "preceding": unitary_block_index
                    }

                    # for this qubit, also add this unitary block as succeeding, if there is
                    # a special gate previously encountered on this qubit
                    if not isinstance(most_recent_special_gate_per_qubit[qubit][0], I):
                        dict_to_update = _get_relevant_dict_to_update(
                            most_recent_special_gate_per_qubit[qubit][0],
                            reset_dict,
                            meas_dict,
                            two_q_dict,
                        )
                        dict_to_update[
                            (
                                most_recent_special_gate_per_qubit[qubit][1],
                                qubit,
                                most_recent_special_gate_per_qubit[qubit][
                                    0
                                ].stim_string,
                            )
                        ]["succeeding"] = unitary_block_index
                    most_recent_special_gate_per_qubit[qubit] = (gate, layer_i)

                    # reset current gates per qubit, update index
                    current_gates_per_qubit[qubit] = []
                    unitary_block_index += 1
        else:
            non_gate_layers[layer_i] = layer

    # after reaching end of circuit, there may be some trailing unitaries that
    # need to be added as `succeeding` blocks.
    # iterate over all qubits to ensure none are missed
    for qubit in circuit.qubits:
        unitary_blocks[unitary_block_index] = current_gates_per_qubit[qubit]
        if not isinstance(most_recent_special_gate_per_qubit[qubit][0], I):
            dict_to_update = _get_relevant_dict_to_update(
                most_recent_special_gate_per_qubit[qubit][0],
                reset_dict,
                meas_dict,
                two_q_dict,
            )
            dict_to_update[
                (
                    most_recent_special_gate_per_qubit[qubit][1],
                    qubit,
                    most_recent_special_gate_per_qubit[qubit][0].stim_string,
                )
            ]["succeeding"] = unitary_block_index
        else:
            # this qubit had no meas, reset or 2q gates applied to it
            continue
        unitary_block_index += 1

    return CompilationData(
        unitary_blocks,
        reset_dict,
        meas_dict,
        two_q_dict,
        non_gate_layers,
        circuit.iterations,
        len(circuit.layers),
    )


def _add_gates_from_unitary_blocks(
    unitary_blocks: dict[int, list[Gate]],
    gate_unitary_block_info: dict[str, int],
    circuit_layers: defaultdict[int, GateLayer],
    layer_index: int,
    unitary_blocks_already_added: set[int],
):
    r"""
    Edit the given `circuit_layers` by adding gates from the unitary blocks to the relevant
    `GateLayer`\ s.

    Parameters
    ----------
    unitary_blocks : Dict[int, List[Gate]]
        Dictionary mapping unitary block indices to lists of single-qubit gates.
    gate_unitary_block_info : Dict[str, int]
        For the given special gate, a dictionary describing which unitary blocks
        precede and succeed it.
    circuit_layers : DefaultDict[int, GateLayer]
        DefaultDictionary storing the GateLayers of the circuit to be constructed.
        Will be edited by adding gates to GateLayers as per the other parameters.
    layer_index : int
        Current layer index, to know which layer indices to add gates to preceding
        and succeeding the special gate.
    unitary_blocks_already_added : Set[int]
        Set to look up which unitary blocks have already been added, so we do not
        add gates more than once.
    """
    # iterate backwards through the preceding block and put in preceding gate layers
    if gate_unitary_block_info["preceding"] not in unitary_blocks_already_added:
        preceding_unitary_block = unitary_blocks[gate_unitary_block_info["preceding"]]
        for i in range(len(preceding_unitary_block)):
            circuit_layers[layer_index - (i + 1)].add_gates(
                preceding_unitary_block[::-1][i]
            )

    # iterate forwards through the succeeding block and put in succeeding gate layers
    if gate_unitary_block_info["succeeding"] not in unitary_blocks_already_added:
        succeeding_unitary_block = unitary_blocks[gate_unitary_block_info["succeeding"]]
        for i, gate in enumerate(succeeding_unitary_block):
            circuit_layers[layer_index + (i + 1)].add_gates(gate)


def _create_circuit_from_compilation_data(
    comp_data: CompilationData, layer_index_lookup: dict[int, int], iterations: int = 1
) -> Circuit:
    """
    The inverse of `_extract_structure_from_circuit` - from a CompilationData object,
    create a `deltakit.circuit.Circuit`.

    Parameters
    ----------
    comp_data : CompilationData
        CompilationData from which to construct a circuit.
    layer_index_lookup : Dict[int, int]
        Dictionary with keys of the original layer indices, and values being their adjusted
        indices accounting for the addition of unitary gates during the compilation process.
    iterations : int
        The number of iterations of the created circuit.

    Returns
    -------
    Circuit
        `deltakit.circuit.Circuit` representation of the circuit described by CompilationData.
    """
    circuit_layers: defaultdict[int, GateLayer] = defaultdict(GateLayer)
    unitary_blocks_already_added: set[int] = set()

    # insert one-qubit special gates
    for special_gate_dict, gate_string_lookup in (
        (comp_data.reset_gates, RESET_GATE_MAPPING),
        (comp_data.measurement_gates, MEASUREMENT_GATE_MAPPING),
    ):
        for gate_info, unitary_block_info in special_gate_dict.items():
            # add special gate to relevant layer
            layer_index, qubit, gate = (
                layer_index_lookup[gate_info[0]],
                gate_info[1],
                gate_info[2],
            )
            circuit_layers[layer_index].add_gates(gate_string_lookup[gate](qubit))

            _add_gates_from_unitary_blocks(
                comp_data.unitary_blocks,
                unitary_block_info,
                circuit_layers,
                layer_index,
                unitary_blocks_already_added,
            )
            unitary_blocks_already_added.add(unitary_block_info["preceding"])
            unitary_blocks_already_added.add(unitary_block_info["succeeding"])

    # insert two-qubit gates
    two_q_gate_dict_as_list = list(comp_data.two_qubit_gates.items())
    for i in range(0, len(two_q_gate_dict_as_list), 2):
        # iterate in pairs
        qubit1_gate_info, qubit1_unitary_block_info = two_q_gate_dict_as_list[i]
        qubit2_gate_info, qubit2_unitary_block_info = two_q_gate_dict_as_list[i + 1]
        layer_index, two_q_gate = (
            layer_index_lookup[qubit1_gate_info[0]],
            TWO_QUBIT_GATE_MAPPING[qubit1_gate_info[2]],
        )
        circuit_layers[layer_index].add_gates(
            two_q_gate(qubit1_gate_info[1], qubit2_gate_info[1])
        )

        # insert unitary blocks
        for unitary_block_info in (
            qubit1_unitary_block_info,
            qubit2_unitary_block_info,
        ):
            _add_gates_from_unitary_blocks(
                comp_data.unitary_blocks,
                unitary_block_info,
                circuit_layers,
                layer_index,
                unitary_blocks_already_added,
            )
            unitary_blocks_already_added.add(unitary_block_info["preceding"])
            unitary_blocks_already_added.add(unitary_block_info["succeeding"])

    # add non-GateLayers
    for layer_index, layer in comp_data.non_gatelayer_layers.items():
        circuit_layers[layer_index_lookup[layer_index]] = layer

    # need to sort circuit_layers, or we may get layers out of order
    return Circuit(dict(sorted(circuit_layers.items())).values(), iterations=iterations)



# (c) Copyright Riverlane 2020-2025.
"""
This module provides a function which constructs a `deltakit.circuit.Circuit`
from a sequence of stages.
"""

from __future__ import annotations

import itertools
from collections.abc import Sequence

from deltakit_circuit import Circuit

from deltakit_explorer.codes._css._css_stage import CSSStage
from deltakit_explorer.codes._css._detectors import get_stage_transition_circuit
from deltakit_explorer.qpu._circuits import merge_layers


def experiment_circuit(experiment: Sequence[CSSStage]) -> Circuit:
    """
    Return a noiseless Circuit for an experiment represented by the input
    Sequence of CSSStage instances.

    Parameters
    ----------
    experiment : Sequence[CSSStage]
        Sequence of CSSStage instances that defines the experiment.

    Returns
    -------
    Circuit
        Noiseless experiment circuit.

    Raises
    ------
    ValueError
        If experiment contains fewer than 3 CSSStage instances.
    ValueError
        If experiment doesn't start with a CSSStage which consists only of qubit
        resets.
    ValueError
        If experiment doesn't end with a CSSStage which consists only of qubit
        measurements and observable additions.
    """
    if len(experiment) < 3:
        msg = "Experiment should contain at least three stages."
        raise ValueError(msg)
    if not experiment[0].resets_only:
        raise ValueError(
            "Experiment should start with a CSSStage which"
            + " consists only of qubit resets."
        )
    if not experiment[-1].allowable_final_stage:
        msg = (
            "Experiment should end with a CSSStage with properties as described in "
            "allowable_final_stage."
        )
        raise ValueError(msg)
    circuit = experiment[0].remaining_rounds

    # Get the detectors for stage transitions.
    stage_transition_detectors = [
        get_stage_transition_circuit(previous_stage, current_stage)
        for previous_stage, current_stage in itertools.pairwise(experiment)
    ]
    for stage, stage_transition_detector in zip(
        experiment[1:], stage_transition_detectors
    ):
        # Apply first round with the observables
        circuit.append_layers(stage.first_round)
        # Apply transition detectors
        circuit.append_layers(stage_transition_detector)
        # Apply remaining rounds
        circuit.append_layers(stage.remaining_rounds)

    # Compress circuit
    return merge_layers(circuit, break_repeat_blocks=True)



# (c) Copyright Riverlane 2020-2025.
"""Exception classes for the project"""
from __future__ import annotations


class ServerException(Exception):  # pragma: nocover
    """Exception, which happened on a server side."""

    message: str
    """Error message from server"""

    def __init__(self, message: str):
        self.message = message



# (c) Copyright Riverlane 2020-2025.
"""This file contains decorators which guard client's input.
Using decorators is motivated by separation of logic with parameter
validation.
"""
from __future__ import annotations

import functools
import itertools
from typing import Any

import stim
import tqdm

from deltakit_explorer._utils._logging import Logging
from deltakit_explorer._utils._utils import HTTP_PACKET_LIMIT
from deltakit_explorer.data._data_analysis import (
    get_decoding_request_size_estimate as get_dec_request_size,
)
from deltakit_explorer.data._data_analysis import (
    get_decoding_response_size_estimate as get_dec_size,
)
from deltakit_explorer.data._data_analysis import (
    get_simulation_response_size_estimate as get_sim_size,
)
from deltakit_explorer.data._data_analysis import has_leakage
from deltakit_explorer.enums._api_enums import DataFormat, DecoderType, QECECodeType
from deltakit_explorer.types._exceptions import ServerException
from deltakit_explorer.types._experiment_types import QECExperimentDefinition
from deltakit_explorer.types._types import (
    CircuitParameters,
    Decoder,
    DecodingResult,
    DetectionEvents,
    LeakageFlags,
    Measurements,
    ObservableFlips,
)


def validate_and_split_decoding(func):
    """Decorator which validates parameters for decoding,
    And warns about potential high consumption."""

    # pylint: disable=too-many-arguments,too-many-positional-arguments,too-many-locals
    @functools.wraps(func)
    def wrapper(
        obj,
        detectors: DetectionEvents,
        observables: ObservableFlips,
        decoder: Decoder,
        noisy_stim_circuit: str | stim.Circuit,
        leakage_flags: LeakageFlags | None = None,
    ):
        if decoder.decoder_type == DecoderType.LCD and leakage_flags is not None:
            Logging.warn(
                "Leakage-aware decoding has a heavy initialisation part. "
                "Big tasks may be cancelled by server timeout.",
                uid="decorator"
            )

        numpy_detectors = detectors.as_numpy()
        # numpy-based instance, to avoid multiple recreation
        detectors = DetectionEvents(
            numpy_detectors,
            DataFormat.B8,
            data_width=numpy_detectors.shape[1]
        )
        shots = numpy_detectors.shape[0]
        batch_size = min(100_000, shots)
        if decoder.decoder_type == DecoderType.LCD:  # LCD weakly avoids batching
            batch_size = shots
        while (
            get_dec_size(noisy_stim_circuit, batch_size, DataFormat.F01)
            > HTTP_PACKET_LIMIT
        ) and batch_size > 100:
            batch_size //= 2
        while (
            get_dec_request_size(noisy_stim_circuit, batch_size, DataFormat.B8)
            > HTTP_PACKET_LIMIT
        ) and batch_size > 100:
            batch_size //= 2
        batch_size = max(1, batch_size)
        if (
            decoder.decoder_type in {DecoderType.AC, DecoderType.BP_OSD}
            and len(str(noisy_stim_circuit)) > 25_000  # e.g. toric 5x5x5
            and numpy_detectors.shape[0] > 10_000  # shots
        ):
            Logging.warn(
                "If you decode with a hypergraph-decoder (AC, BP-OSD), "
                "decoding time may be significant and subject to the "
                "server timeout.",
                uid="decorator",
            )
        # batch decoding
        batch_detectors = detectors.to_batches(batch_size)
        batch_observables = observables.to_batches(batch_size)
        if leakage_flags is not None:
            batch_leakage = leakage_flags.to_batches(batch_size)
        else:
            batch_leakage = []
        results = []
        tasks: tqdm.tqdm[Any] | itertools.zip_longest[Any] = itertools.zip_longest(
            batch_detectors, batch_observables, batch_leakage
        )
        if batch_size < shots:
            iterations = (shots + batch_size - 1) // batch_size
            tasks = tqdm.tqdm(tasks, "decoding batches", total=iterations)

        for dets, obs, leakage in tasks:
            decoding_result = func(
                obj, dets, obs,
                decoder, noisy_stim_circuit, leakage,
            )
            results.append(decoding_result)
        return DecodingResult.combine(results)

    return wrapper


def validate_generation(func):
    """Decorator which validates circuit generation and
    warns if the task may consume a lot of server resources,
    which may lead to a timeout."""

    @functools.wraps(func)
    def wrapper(
        obj, experiment_definition: QECExperimentDefinition,
    ):
        parameters: CircuitParameters | None = experiment_definition.parameters
        if experiment_definition.code_type in {
            QECECodeType.ROTATED_PLANAR,
            QECECodeType.UNROTATED_PLANAR,
            QECECodeType.UNROTATED_TORIC
        } and parameters is not None and parameters.sizes is not None:
            area = functools.reduce(
                lambda a, b: a * b,
                parameters.sizes.sizes, 1)
            if area >= 21 ** 2 and experiment_definition.basis_gates is not None:
                Logging.warn(
                    "Circuit generation with a provided gate set may be slow for "
                    "big code patches. This may lead to a server timeout.",
                    uid="decorator",
                )
        if (
            experiment_definition.code_type == QECECodeType.BIVARIATE_BICYCLE
            and parameters is not None
            and parameters.matrix_specifications is not None
        ):
            param_l = parameters.matrix_specifications.param_l
            has_basis_gates = experiment_definition.basis_gates is not None
            if param_l >= 15 and has_basis_gates:
                Logging.warn(
                    "BBCode circuit generation with a provided gate set "
                    "may be slow. This may lead to a server timeout.",
                    uid="decorator",
                )
        return func(obj, experiment_definition)

    return wrapper


def _split_into_batches(total: int, batch_size: int) -> list[int]:
    shots_list = []
    while total > 0:
        shots_list.append(min(batch_size, total))
        total -= batch_size
    return shots_list


def validate_and_split_simulation(func):
    """Decorator which validates parameters for simulation,
    And warns about potential high resource consumption."""

    @functools.wraps(func)
    def wrapper(  # pragma: nocover
        obj,
        stim_circuit: str | stim.Circuit,
        shots: int,
    ):
        assert shots > 0, "Number of shots should be positive."
        batch_size = min(100_000, shots)  # initial guess
        while (
            get_sim_size(stim_circuit, batch_size, DataFormat.F01) > HTTP_PACKET_LIMIT
            and batch_size > 100
        ):
            batch_size //= 2
        batch_size = max(1, batch_size)

        # do not split if simulated locally
        if not has_leakage(str(stim_circuit)):
            batch_size = shots

        shots_list: list[int] | tqdm.tqdm[int] = _split_into_batches(shots, batch_size)

        if len(shots_list) > 1:
            Logging.warn(
                "Simulation will be performed in batches "
                f"of {batch_size} shots.",
                uid="decorator",
            )
        measurements = []
        leakages = []
        if len(shots_list) > 1:
            shots_list = tqdm.tqdm(shots_list, "simulation batches")
        for batch in shots_list:
            Logging.info(f"Simulating {batch} shots...", uid="decorator")
            meas, leak = func(obj, stim_circuit, batch)
            measurements.append(meas)
            leakages.append(leak)
        if (
            not all(leak is None for leak in leakages)
            and not all(leak is not None for leak in leakages)
        ):
            msg = (
                "Some of batches returned leakage information, "
                "and some did not. Data is inconsistent."
            )
            raise ServerException(
                msg
            )
        final_measurements = Measurements.combine(measurements)
        final_leakages = None
        if all(leak is not None for leak in leakages):
            final_leakages = LeakageFlags.combine(leakages)
        return (final_measurements, final_leakages)

    return wrapper



import warnings
from collections.abc import Mapping, Sequence

import numpy as np
import numpy.typing as npt
import pandas as pd

from deltakit_explorer.analysis import (
    LogicalErrorProbabilityPerRoundResults,
    compute_logical_error_per_round,
)
from deltakit_explorer.analysis._analysis import calculate_lep_and_lep_stddev
from deltakit_explorer.analysis._lambda import (
    LambdaResults,
    calculate_lambda_and_lambda_stddev,
)


def _filter_non_close_noise_parameters(
    data: pd.DataFrame,
    noise_parameters: npt.NDArray[np.floating],
    noise_parameter_names: Sequence[str],
) -> pd.DataFrame:
    """Return a filtered view over ``data`` such that all entries are close to the
    provided ``noise_parameters``."""
    ret = data
    for name, param in zip(noise_parameter_names, noise_parameters):
        ret = ret[np.isclose(ret[f"noise_{name}"], param)]
    return ret


def compute_lambda_and_stddev_from_results(
    xi: npt.NDArray[np.floating],
    noise_parameter_names: Sequence[str],
    num_rounds_by_distance: Mapping[int, Sequence[int]],
    data: pd.DataFrame,
) -> tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]]:
    """Compute Λ from ``data`` for all the provided noise parameters in ``xi``.

    This function assumes that the provided ``data`` has been generated with
    ``RunAllAnalysisEngine`` and the tasks returned by
    ``generate_decoder_managers_for_lambda`` called on the exact same parameters as
    provided here.

    Args:
        xi (npt.NDArray[numpy.floating]): an array of shape ``(m, n)``, where ``m`` is
            the number of noise parameters needed by ``noise_model_type`` and ``n`` the
            number of points to evaluate Λ on.
        num_rounds_by_distances (Mapping[int, Sequence[int]]): the different number of
            rounds that will be used to estimate the logical error-rate per round for
            each code distance.
        data (pandas.DataFrame): data generated by executing ``RunAllAnalysisEngine`` on
            the tasks returned by ``generate_decoder_managers_for_lambda`` called
            with the exact same parameters as provided to this function.

    Returns:
        two arrays of dimensions ``(1, n)``. The first array contains an estimation of Λ
        for each parameters in ``xi`` and the second array contains the computed
        standard deviations.
    """
    if len(xi.shape) != 2:
        msg = f"Expected a 2-dimensional array but got shape {xi.shape}."
        raise ValueError(msg)
    _, n = xi.shape
    ret: npt.NDArray[np.floating] = np.zeros((1, n), dtype=np.float64)
    stddev: npt.NDArray[np.floating] = np.zeros_like(ret)
    for i in range(n):
        filtered_data = _filter_non_close_noise_parameters(
            data, xi[:, i], noise_parameter_names
        )
        res = _compute_lambda_from_results(num_rounds_by_distance, filtered_data)
        ret[0, i], stddev[0, i] = res.lambda_, res.lambda_stddev
    return ret, stddev


def _compute_lambda_from_results(
    num_rounds_by_distance: Mapping[int, Sequence[int]],
    data: pd.DataFrame,
) -> LambdaResults:
    lepprs: list[float] = []
    leppr_stddevs: list[float] = []
    distances = sorted(num_rounds_by_distance.keys())
    for d in distances:
        leppr = _compute_logical_error_rate_per_round_from_results(
            num_rounds_by_distance[d], data[data["distance"] == d]
        )
        lepprs.append(leppr.leppr)
        leppr_stddevs.append(leppr.leppr_stddev)
    return calculate_lambda_and_lambda_stddev(distances, lepprs, leppr_stddevs)


def _compute_logical_error_rate_per_round_from_results(
    num_rounds: npt.NDArray[np.int_] | Sequence[int], data: pd.DataFrame
) -> LogicalErrorProbabilityPerRoundResults:
    num_fails: npt.NDArray[np.int_] | list[int] = []
    max_shots: npt.NDArray[np.int_] | list[int] = []
    for nrounds in num_rounds:
        data_row = data.query(f"num_rounds == {nrounds}")
        nfails = data_row["fails"].to_numpy()[0]
        nshots = data_row["shots"].to_numpy()[0]
        num_fails.append(nfails)
        max_shots.append(nshots)
    # Filter out 0 fails
    non_zeros_mask = np.asarray(num_fails) != 0
    if not np.all(non_zeros_mask):
        warnings.warn(
            "Found at least one sampling task with 0 fails. It will be ignored for "
            "logical error probability computation."
        )
    num_rounds = np.asarray(num_rounds)[non_zeros_mask]
    num_fails = np.asarray(num_fails)[non_zeros_mask]
    max_shots = np.asarray(max_shots)[non_zeros_mask]

    lep, lep_stddev = calculate_lep_and_lep_stddev(num_fails, max_shots)
    return compute_logical_error_per_round(num_rounds, lep, lep_stddev)



from collections.abc import Callable, Sequence

from deltakit_circuit import Circuit, Qubit
from deltakit_circuit.gates import (
    CXSWAP,
    CZSWAP,
    ISWAP,
    ISWAP_DAG,
    MEASUREMENT_GATE_MAPPING,
    ONE_QUBIT_GATE_MAPPING,
    RESET_GATE_MAPPING,
    TWO_QUBIT_GATE_MAPPING,
    Gate,
    SymmetricTwoQubitGate,
)

from deltakit_explorer.qpu._circuits._circuit_functions import merge_layers
from deltakit_explorer.qpu._native_gate_set import (
    NativeGateSetAndTimes,
    OneQubitCliffordGate,
    TwoOperandGate,
)

from ._tableau_functions import (
    CZ_TO_GATE_DICT,
    CZSWAP_TO_GATE_DICT,
    GATE_TO_CZ_DICT,
    GATE_TO_CZSWAP_DICT,
    MEAS_COMPILATION_LOOKUP_DICT,
    RESET_COMPILATION_LOOKUP_DICT,
    CompilationData,
    EquivalentTableauDict,
    SpecialGateDict,
    TableauDict,
    TwoQubitGateCompilationLookupDict,
    TwoQubitGateDictEntry,
    _create_circuit_from_compilation_data,
    _extract_structure_from_circuit,
    _get_compilation_dict,
    _get_compilation_with_measurement_after_unitaries,
    _get_compilation_with_projectors_before_unitaries,
    _get_compilation_with_two_qubit_gates,
    _get_tableau_as_key,
    _get_tableau_from_sequence_of_1q_gates,
    _get_tableau_key_from_sequence_of_gates,
)


def _compile_to_native_gates_plus_unitaries(
    gate_info: tuple[int, Qubit, str],
    preceding_ub: list[OneQubitCliffordGate],
    succeeding_ub: list[OneQubitCliffordGate],
    available_native_gates: set[Gate],
    available_single_qubit_gates: set[OneQubitCliffordGate],
    compilation_lookup_dict: dict[
        Gate, dict[Gate, tuple[list[OneQubitCliffordGate], list[OneQubitCliffordGate]]]
    ],
    special_gate_mapping: dict[str, Gate],
    comp_dict: TableauDict,
    up_to_paulis: bool,
) -> tuple[list[OneQubitCliffordGate], list[OneQubitCliffordGate], str]:
    """
    Given a reset or measurement gate and its surrounding unitaries, attempt to compile it
    to the native gate set, by trying each available reset/measurement gate and trying to
    compile the resulting unitaries into native gates. If this fails, a ValueError is raised.

    Parameters
    ----------
    gate_info : Tuple[int, Qubit, str]
        Current gate's info, from the special gate dictionary, detailing the layer index,
        qubit and stim_string of the gate.
    preceding_ub : List[OneQubitCliffordGate]
        Unitary block preceding the gate.
    succeeding_ub : List[OneQubitCliffordGate]
        Unitary block succeeding the gate.
    available_native_gates : Set[Gate]
        Available native reset/measurement gates to try to compile to.
    available_single_qubit_gates : Set[OneQubitCliffordGate]
        Available single qubit gates to compile the unitary blocks to.
    compilation_lookup_dict : Dict[
        Gate, Dict[Gate, Tuple[List[OneQubitCliffordGate], List[OneQubitCliffordGate]]]
    ]
        Dictionary detailing how to compile from one reset/measurement gate to another.
    special_gate_mapping : Dict[str, Gate]
        Dictionary to lookup the stim_string to gate type.
    comp_dict : TableauDict
        Compilation dictionary, with tableaus as keys and lists of unitary
        gates as values.
    up_to_paulis : bool
        Whether to compute compilation up to pauli terms or not.

    Returns
    -------
    Tuple[List[OneQubitCliffordGate], List[OneQubitCliffordGate], str]
        A tuple containing the preceding and succeeding unitary blocks after
        compilation to native gates, and the stim_string of the gate compiled to.
    """
    # try compiling to a native gate. If we run out, the compilation is not
    # possible, so throw an error.
    while len(available_native_gates) >= 0:
        try:
            target_gate = available_native_gates.pop()
        except KeyError as ke:
            msg = (
                "Unable to compile to provided native reset and measurement gates,"
                " please try changing the native gate set."
            )
            raise ValueError(msg) from ke

        # look up the compilation from current gate to target native gate.
        # if the entry is not in the dictionary, the compilation is not
        # supported, so try the next native gate.
        try:
            unitaries_to_add = compilation_lookup_dict[
                special_gate_mapping[gate_info[2]]
            ][target_gate]
        except KeyError:
            continue

        updated_preceding_ub, updated_succeeding_ub = [], []
        for unitary_block, ub_is_preceding in (
            (preceding_ub, True),
            (succeeding_ub, False),
        ):
            # try compiling the updated unitary_block to native gates. if any contain
            # a non-native gate after compilation, stop and try the next target gate.
            # otherwise, stop, as compilation succeeded
            if ub_is_preceding:
                updated_preceding_ub = _compile_or_exchange_unitary_block(
                    unitary_block + [g(gate_info[1]) for g in unitaries_to_add[0]],
                    comp_dict,
                    up_to_paulis,
                )
            else:
                updated_succeeding_ub = _compile_or_exchange_unitary_block(
                    [g(gate_info[1]) for g in unitaries_to_add[1]] + unitary_block,
                    comp_dict,
                    up_to_paulis,
                )

        # if the gates in the unitary blocks are a subset of the native gates, we're done
        if set({type(g) for g in updated_preceding_ub + updated_succeeding_ub}) <= set(
            available_single_qubit_gates
        ):
            break
    return updated_preceding_ub, updated_succeeding_ub, target_gate.stim_string


def _compile_reset_to_native_gates_plus_unitaries(
    gate_info: tuple[int, Qubit, str],
    preceding_ub: list[OneQubitCliffordGate],
    succeeding_ub: list[OneQubitCliffordGate],
    native_gate_set: NativeGateSetAndTimes,
    comp_dict: TableauDict,
    up_to_paulis: bool,
) -> tuple[list[OneQubitCliffordGate], list[OneQubitCliffordGate], str]:
    return _compile_to_native_gates_plus_unitaries(
        gate_info,
        preceding_ub,
        succeeding_ub,
        set(native_gate_set.reset_gates.keys()),
        set(native_gate_set.one_qubit_gates.keys()),
        RESET_COMPILATION_LOOKUP_DICT,
        RESET_GATE_MAPPING,
        comp_dict,
        up_to_paulis,
    )


def _compile_measurement_to_native_gates_plus_unitaries(
    gate_info: tuple[int, Qubit, str],
    preceding_ub: list[OneQubitCliffordGate],
    succeeding_ub: list[OneQubitCliffordGate],
    native_gate_set: NativeGateSetAndTimes,
    comp_dict: TableauDict,
    up_to_paulis: bool,
) -> tuple[list[OneQubitCliffordGate], list[OneQubitCliffordGate], str]:
    return _compile_to_native_gates_plus_unitaries(
        gate_info,
        preceding_ub,
        succeeding_ub,
        set(native_gate_set.measurement_gates.keys()),
        set(native_gate_set.one_qubit_gates.keys()),
        MEAS_COMPILATION_LOOKUP_DICT,
        MEASUREMENT_GATE_MAPPING,
        comp_dict,
        up_to_paulis,
    )


def _compile_or_exchange_unitary_block(
    unitary_block: list[OneQubitCliffordGate],
    comp_dict: TableauDict,
    up_to_paulis: bool,
) -> list[OneQubitCliffordGate]:
    """
    Given a unitary block in the form of a list of one qubit gates that are possibly not
    native gates, attempt to compile the unitary block to a shorter set of gates, so
    that the unitary block is expressed in terms of only native gates.

    Parameters
    ----------
    unitary_block : List[OneQubitCliffordGate]
        Unitary block to attempt to shorten or exchange for an equivalent expression in terms
        of just native gates.
    comp_dict : TableauDict
        Dictionary containing the shortest discovered expression for a given tableau.
    up_to_paulis : bool
        Boolean specifying whether to perform the compilation up to pauli terms or not.

    Returns
    -------
    List[OneQubitCliffordGate]
        Updated unitary block.
    """
    # get tableau of unitary block
    ub_tab = _get_tableau_as_key(
        _get_tableau_from_sequence_of_1q_gates([g.stim_string for g in unitary_block]),
        up_to_paulis=up_to_paulis,
    )

    # compile or exchange
    if ub_tab in comp_dict:
        # if we can compile to something shorter, do that
        unitary_block = [
            ONE_QUBIT_GATE_MAPPING[g](unitary_block[0].qubit) for g in comp_dict[ub_tab]
        ]
    return unitary_block


def _compile_reset_and_meas_to_native_gates(
    comp_data: CompilationData,
    native_gate_set: NativeGateSetAndTimes,
    comp_dict: TableauDict,
    up_to_paulis: bool,
    layer_index_lookup: dict[int, int],
) -> tuple[CompilationData, dict[int, int]]:
    """
    Given a circuit in the form of CompilationData, attempt to compile the
    reset and measurement gates into native gates, possibly with extra unitary
    gates inserted into the surrounding unitary blocks. We then attempt to compile
    or exchange the unitary blocks for something in terms of native gates. If this
    is not possible, the compilation has failed, so we throw a ValueError.
    After the unitary blocks have been updated, layer_index_lookup is modified
    to allow room for any extra unitaries that have been added. For instance,
    when creating CompilationData, say a reset and two qubit gate are separated
    by one gate layer in between.

    If we then compiled the reset and consequently require two unitaries following
    the reset, we need to add another layer in between the reset and two qubit gate.
    This information is tracked in layer_index_lookup, where the positive delta of
    unitary blocks before and after compilation is added to the succeeding layer
    indices to make room for the new unitaries. In the example mentioned above,
    if the reset was at layer 10, and the two qubit gate at layer 12, after
    compilation the two qubit gate would be at layer 13, and the dictionary would
    have entry {12: 13} to reflect this. Then when the circuit is reconstructed from
    CompilationData, the layer indices recorded before compilation are looked up in
    the layer_index_lookup dictionary to add the gates at the right layer indices
    such that there is room for the unitaries.

    Parameters
    ----------
    comp_data : CompilationData
        Circuit to compile in the form of CompilationData.
    native_gate_set : NativeGateSetAndTimes
        Native gate set to compile to.
    comp_dict : TableauDict
        Dictionary containing the shortest discovered expression for a given tableau.
    up_to_paulis : bool
        Boolean specifying whether to perform the compilation up to pauli terms or not.
    layer_index_lookup : Dict[int, int]
        Dictionary with keys of the original layer indices, and values being their adjusted
        indices accounting for the addition of unitary gates during the compilation process.

    Returns
    -------
    Tuple[CompilationData, Dict[int, int]]
        A tuple containing the compiled circuit in the form of CompilationData,
        and the index lookup dictionary that has as keys the original layer index
        and as values the adjusted index of that layer after taking into account
        the addition of unitary gates in between layers due to compilation.
    """
    # Takes the same form of SpecialGateDict, but will be updated
    # to reflect the compilation. E.g, originally we may have
    # (0, Qubit(0), "RX") for an RX on qubit 0 in the first gate layer,
    # but after compiling to RZ, this will change to (0, Qubit(0), "RZ").
    # We make new dictionaries because dictionary keys are immutable.
    compiled_reset_gates: SpecialGateDict = {}
    compiled_measurement_gates: SpecialGateDict = {}
    for special_dict, dict_to_update, compile_function, gate_lookup, native_gates in (
        (
            comp_data.reset_gates,
            compiled_reset_gates,
            _compile_reset_to_native_gates_plus_unitaries,
            RESET_GATE_MAPPING,
            set(native_gate_set.reset_gates.keys()),
        ),
        (
            comp_data.measurement_gates,
            compiled_measurement_gates,
            _compile_measurement_to_native_gates_plus_unitaries,
            MEASUREMENT_GATE_MAPPING,
            set(native_gate_set.measurement_gates.keys()),
        ),
    ):
        for gate_info, gate_ubs in special_dict.items():
            if gate_lookup[gate_info[2]] in native_gates:
                dict_to_update[gate_info] = gate_ubs
                continue

            # Try compiling to the current target native gate by seeing how the
            # surrounding unitary blocks change
            ub1_index, ub2_index = gate_ubs["preceding"], gate_ubs["succeeding"]
            ub1, ub2, compiled_gate = compile_function(
                gate_info,
                comp_data.unitary_blocks[ub1_index],
                comp_data.unitary_blocks[ub2_index],
                native_gate_set,
                comp_dict,
                up_to_paulis,
            )
            # check whether the preceding or succeeding unitaries have got longer
            # than in the original circuit, in which case we must update the layer indices
            # of special gates following the current special gate
            for ub_index, new_ub, start_delta in (
                (ub1_index, ub1, 0),
                (ub2_index, ub2, 1),
            ):
                # start_delta is 0 when we consider the preceding unitary block, so
                # that we adjust the layer index of the current gate. It is then 1
                # for the succeeding, since we do not need to adjust the layer index
                # of the current gate if only the succeeding unitaries have gotten longer
                delta = len(new_ub) - len(comp_data.unitary_blocks[ub_index])
                if delta > 0:
                    for i in range(
                        gate_info[0] + start_delta, max(layer_index_lookup.keys()) + 1
                    ):
                        layer_index_lookup[i] += delta

            # update relevant parts of comp_data
            comp_data.unitary_blocks[ub1_index] = ub1
            comp_data.unitary_blocks[ub2_index] = ub2
            dict_to_update[(gate_info[0], gate_info[1], compiled_gate)] = gate_ubs

    comp_data.reset_gates = compiled_reset_gates
    comp_data.measurement_gates = compiled_measurement_gates
    return comp_data, layer_index_lookup


def _compile_two_qubit_gate_to_target(
    current_gate: TwoOperandGate,
    target_gate: TwoOperandGate,
    qubit1_entry: TwoQubitGateDictEntry,
    qubit2_entry: TwoQubitGateDictEntry,
    gate_to_intermediate_rep_dict: TwoQubitGateCompilationLookupDict,
    int_rep_to_target_dict: TwoQubitGateCompilationLookupDict,
    unitary_blocks: dict[int, list],
) -> tuple[
    list[OneQubitCliffordGate],
    list[OneQubitCliffordGate],
    list[OneQubitCliffordGate],
    list[OneQubitCliffordGate],
]:
    r"""
    Given a two-qubit gate, and a target two-qubit gate, compute the updated
    unitary blocks required to compile the current gate to the target gate.
    This is done by first working out the expression of `current_gate` in
    terms of the intermediate representation (CZ or CZSWAP) + unitaries,
    then going from the intermediate representation to `target_gate` with
    further unitaries.

    Parameters
    ----------
    current_gate : TwoOperandGate
        The current gate, to be compiled from.
    target_gate : TwoOperandGate
        The target gate, to be compiled to. I.e, `current_gate` will be
        expressed in terms of `target_gate`.
    qubit1_entry : TwoQubitGateDictEntry
        The entry of CompilationData.two_qubit_gates corresponding to
        qubit 1 of the two qubit gate.
    qubit2_entry : TwoQubitGateDictEntry
        The entry of CompilationData.two_qubit_gates corresponding to
        qubit 2 of the two qubit gate.
    gate_to_int_rep_dict : TwoQubitGateCompilationLookupDict
        Dictionary providing lookup from `current_gate` to the intermediate
        representation (CZ or CZSWAP) for compiling.
        Should be either GATE_TO_CZ_DICT or GATE_TO_CZSWAP_DICT.
    int_rep_to_target_dict : TwoQubitGateCompilationLookupDict
        Dictionary providing lookup from the intermediate
        representation (CZ or CZSWAP) to `target_gate`.
        Should be either CZ_TO_GATE_DICT or CZSWAP_TO_GATE_DICT.
    unitary_blocks : Dict[int, List]
        The CompilationData.unitary_blocks dictionary, providing information
        on the indices of unitary blocks and their contents.

    Returns
    -------
    Tuple[
        List[OneQubitCliffordGate],
        List[OneQubitCliffordGate],
        List[OneQubitCliffordGate],
        List[OneQubitCliffordGate],
    ]
        A tuple containing, in order,
        the unitary blocks corresponding to:
                [0]: Before qubit 1
                [1]: After qubit 1
                [2]: Before qubit 2
                [3]: After qubit 2
        Each in the form of a list of `OneQubitCliffordGate`\ s.
    """
    qubit1_info, qubit1_ubs = qubit1_entry
    qubit2_info, qubit2_ubs = qubit2_entry
    ub0, ub1, ub2, ub3 = (
        unitary_blocks[qubit1_ubs["preceding"]],
        unitary_blocks[qubit1_ubs["succeeding"]],
        unitary_blocks[qubit2_ubs["preceding"]],
        unitary_blocks[qubit2_ubs["succeeding"]],
    )

    # First, replace current gate with intermediate representation by
    # adding appropriate unitaries
    try:
        ubs = gate_to_intermediate_rep_dict[current_gate]
    except KeyError as ke:
        msg = "Current gate not present in gate_to_intermediate_rep_dict dictionary"
        raise ValueError(msg) from ke
    ub0 = ub0 + [g(qubit1_info[1]) for g in ubs[0]]
    ub1 = [g(qubit1_info[1]) for g in ubs[1]] + ub1
    ub2 = ub2 + [g(qubit2_info[1]) for g in ubs[2]]
    ub3 = [g(qubit2_info[1]) for g in ubs[3]] + ub3

    # Then, replace the intermediate representation with the target gate by
    # adding appropriate unitaries, taking into account that the intermediate
    # representation gate or this gate may need to be reversed
    try:
        ubs_target = int_rep_to_target_dict[target_gate]
    except KeyError as ke:
        msg = (
            "Cannot compile between groups -"
            f" {current_gate.stim_string} to {target_gate.stim_string} not supported"
        )
        raise ValueError(msg) from ke

    ub0 = ub0 + [g(qubit1_info[1]) for g in ubs_target[0]]
    ub1 = [g(qubit1_info[1]) for g in ubs_target[1]] + ub1
    ub2 = ub2 + [g(qubit2_info[1]) for g in ubs_target[2]]
    ub3 = [g(qubit2_info[1]) for g in ubs_target[3]] + ub3
    return ub0, ub1, ub2, ub3


def _compile_two_qubit_gates_to_native_gates(
    comp_data: CompilationData,
    native_gate_set: NativeGateSetAndTimes,
    comp_dict: TableauDict,
    up_to_paulis: bool,
    layer_index_lookup: dict[int, int],
) -> tuple[CompilationData, dict[int, int]]:
    """
    Given a circuit in the form of CompilationData, attempt to compile the
    two-qubit gates into native gates, possibly with extra unitary
    gates inserted into the surrounding unitary blocks.
    Along the way, `layer_index_lookup` is updated to account for
    extra gate layers needed to fit in the unitaries required for
    computation.
    If compilation to native gates not possible, throws a ValueError.

    Parameters
    ----------
    comp_data : CompilationData
        Circuit to compile in the form of CompilationData.
    native_gate_set : NativeGateSetAndTimes
        Native gate set to compile to.
    comp_dict : TableauDict
        Dictionary containing the shortest discovered expression for a given tableau.
    up_to_paulis : bool
        Boolean specifying whether to perform the compilation up to pauli terms or not.
    layer_index_lookup : Dict[int, int]
        Dictionary with keys of the original layer indices, and values being their adjusted
        indices accounting for the addition of unitary gates during the compilation process.

    Returns
    -------
    Tuple[CompilationData, Dict[int, int]]
        Tuple containing the compiled circuit in the form of CompilationData,
        and the updated `layer_index_lookup`.
    """
    if len(comp_data.two_qubit_gates) == 0:
        return comp_data, layer_index_lookup
    # Updated version of the CompilationData.two_qubit_gates dictionary that will contain
    # the newly compiled versions of each two qubit gate. To be filled in as the compilation
    # proceeds.
    new_two_qubit_dict: SpecialGateDict = {}
    dict_as_list = list(comp_data.two_qubit_gates.items())
    # for each two-qubit gate, attempt to compile it to any of the available native gates
    for i in range(0, len(dict_as_list), 2):
        qubit1_entry: TwoQubitGateDictEntry = dict_as_list[i]
        qubit2_entry: TwoQubitGateDictEntry = dict_as_list[i + 1]
        current_gate = TWO_QUBIT_GATE_MAPPING[qubit1_entry[0][2]]
        if current_gate in native_gate_set.two_qubit_gates:
            # if current gate in native gates, add it as-is to new dict and move on
            new_two_qubit_dict[qubit1_entry[0]] = qubit1_entry[1]
            new_two_qubit_dict[qubit2_entry[0]] = qubit2_entry[1]
            continue

        # set relevant lookup dicts
        if current_gate in (ISWAP, ISWAP_DAG, CXSWAP, CZSWAP):
            gate_to_intm_dict = GATE_TO_CZSWAP_DICT
            intm_to_target_dict = CZSWAP_TO_GATE_DICT
        else:
            gate_to_intm_dict = GATE_TO_CZ_DICT
            intm_to_target_dict = CZ_TO_GATE_DICT

        set_of_possible_gates = set(native_gate_set.two_qubit_gates.keys())
        while len(set_of_possible_gates) >= 0:
            # for each possible native two-qubit gate
            try:
                target_gate = set_of_possible_gates.pop()
            except KeyError as ke:
                msg = (
                    "Unable to compile to provided native two-qubit gates,"
                    " please try changing the native gate set."
                )
                raise ValueError(msg) from ke

            # compile the two qubit gate, by getting the new unitary blocks
            updated_unitaries = _compile_two_qubit_gate_to_target(
                current_gate,
                target_gate,
                qubit1_entry,
                qubit2_entry,
                gate_to_intm_dict,
                intm_to_target_dict,
                comp_data.unitary_blocks,
            )

            # test if the new unitary blocks can be compiled to something in terms of native gates.
            # if not, the compilation failed, so try the next native gate
            compilation_successful = True
            compiled_updated_unitaries = []
            for ub in updated_unitaries:
                compiled_unitary_block = _compile_or_exchange_unitary_block(
                    ub,
                    comp_dict,
                    up_to_paulis,
                )

                # if a non-native gate remains after compilation, cannot compiled to this 2q gate
                if not set({type(g) for g in compiled_unitary_block}) <= set(
                    native_gate_set.one_qubit_gates.keys()
                ):
                    compilation_successful = False
                    break
                compiled_updated_unitaries.append(compiled_unitary_block)

            if compilation_successful:
                # record change in layer indices, based on the unitary block with the
                # highest delta
                ub1_delta = len(compiled_updated_unitaries[0]) - len(
                    comp_data.unitary_blocks[qubit1_entry[1]["preceding"]]
                )
                ub3_delta = len(compiled_updated_unitaries[2]) - len(
                    comp_data.unitary_blocks[qubit2_entry[1]["preceding"]]
                )
                max_delta_before_gate = max(ub1_delta, ub3_delta)

                ub2_delta = len(compiled_updated_unitaries[1]) - len(
                    comp_data.unitary_blocks[qubit1_entry[1]["succeeding"]]
                )
                ub4_delta = len(compiled_updated_unitaries[3]) - len(
                    comp_data.unitary_blocks[qubit2_entry[1]["succeeding"]]
                )
                max_delta_after_gate = max(ub2_delta, ub4_delta)

                # for unitary blocks before, adjust the layer of the gate itself.
                # for unitary blocks after, adjust the layers following the gate, but not the gate itself.
                for start_delta, delta in (
                    (0, max_delta_before_gate),
                    (1, max_delta_after_gate),
                ):
                    if delta > 0:
                        for i in range(
                            qubit1_entry[0][0] + start_delta,
                            max(layer_index_lookup.keys()) + 1,
                        ):
                            layer_index_lookup[i] += delta

                # update unitary blocks
                comp_data.unitary_blocks[
                    qubit1_entry[1]["preceding"]
                ] = compiled_updated_unitaries[0]
                comp_data.unitary_blocks[
                    qubit1_entry[1]["succeeding"]
                ] = compiled_updated_unitaries[1]
                comp_data.unitary_blocks[
                    qubit2_entry[1]["preceding"]
                ] = compiled_updated_unitaries[2]
                comp_data.unitary_blocks[
                    qubit2_entry[1]["succeeding"]
                ] = compiled_updated_unitaries[3]

                new_two_qubit_dict[
                    (
                        qubit1_entry[0][0],
                        qubit1_entry[0][1],
                        target_gate.stim_string,
                    )
                ] = qubit1_entry[1]
                new_two_qubit_dict[
                    (
                        qubit2_entry[0][0],
                        qubit2_entry[0][1],
                        target_gate.stim_string,
                    )
                ] = qubit2_entry[1]
                break
    comp_data.two_qubit_gates = new_two_qubit_dict
    return comp_data, layer_index_lookup


def _compile_comp_data(
    comp_data: CompilationData,
    layer_index_lookup: dict[int, int],
    native_gate_set: NativeGateSetAndTimes,
    comp_dict: TableauDict,
    gate_exchange_dict: EquivalentTableauDict,
    up_to_paulis: bool,
    allow_terms_to_mutate: bool,
    allow_terms_to_multiply: bool,
) -> Circuit:
    """
    Given a CompilationData object, and the corresponding relevant data, compile
    this CompilationData to native gates and return a deltakit_circuit.Circuit.

    Parameters
    ----------
    comp_data : CompilationData
        Circuit to compile, in CompilationData form.
    layer_index_lookup : Dict[int, int]
        Accompanying layer index lookup dictionary, used to ensure
        gates added for compilation do not clash in layers.
    native_gate_set : NativeGateSetAndTimes
        Native gate set to try compiling to.
    comp_dict : TableauDict
        Compilation dictionary containing the shortest discovered gate
        expression for a given tableau.
    gate_exchange_dict : EquivalentTableauDict
        Dictionary containing equivalent expressions in terms of different
        (but still native) gates of each tableau in comp_dict. Used for
        two-qubit gate compilation where equivalent gate expressions
        for a given tableau may give better compilations.
    up_to_paulis: bool
        Specify whether compilation should happen only up to Pauli terms.
        By default, False.
    allow_terms_to_mutate : bool
        For two qubit gate compilation, will allow the case of pulling a single-qubit Pauli
        term through a two-qubit gate such that the resulting (possibly two-qubit) Pauli has
        a different term on either qubit.
        E.g., pulling an X through a CZ from the first qubit mutates into a Z term
        on the second qubit.
        By default, True.
    allow_terms_to_multiply : bool
        For two qubit gate compilation, will allow the case of pulling a term through a
        two-qubit gate such that a second term on the other qubit is created. E.g., pulling
        an X through a CX from the first qubit mutates into an X on the first qubit and an
        X on the second qubit.
        By default, True.

    Returns
    -------
    deltakit_circuit.Circuit
        Circuit compiled from the given CompilationData.
    """
    # compile repeat blocks first - call this function for each CompilationData
    # in the non_gatelayer_layers dict.
    for layer_index, layer in comp_data.non_gatelayer_layers.items():
        if isinstance(layer, CompilationData):
            layer_index_lookup = {
                i: i for i in range(max(comp_data.num_layers, layer.num_layers))
            }
            comp_data.non_gatelayer_layers[layer_index] = _compile_comp_data(
                layer,
                layer_index_lookup,
                native_gate_set,
                comp_dict,
                gate_exchange_dict,
                up_to_paulis,
                allow_terms_to_mutate,
                allow_terms_to_multiply,
            )

    # step 1: compile reset and measurement gates to native reset and
    # measurement gates + some unitaries.
    comp_data, layer_index_lookup = _compile_reset_and_meas_to_native_gates(
        comp_data,
        native_gate_set,
        comp_dict,
        up_to_paulis,
        layer_index_lookup,
    )

    # step 2: compile two-qubit gates to native
    comp_data, layer_index_lookup = _compile_two_qubit_gates_to_native_gates(
        comp_data,
        native_gate_set,
        comp_dict,
        up_to_paulis,
        layer_index_lookup,
    )

    # step 2a: go through all unitary blocks and compile to native, reducing
    # where possible. This is because, in the previous steps,
    # if the unitary blocks preceding a two qubit gate are empty,
    # the following blocks will not be looked at. Similarly, if a reset or measurement
    # is part of the native gates, its unitaries are also ignored.
    for unitary_block_index, unitary_block in comp_data.unitary_blocks.items():
        ub_tab = _get_tableau_as_key(
            _get_tableau_from_sequence_of_1q_gates(
                [g.stim_string for g in unitary_block]
            ),
            up_to_paulis=up_to_paulis,
        )
        if ub_tab in comp_dict and len(comp_dict[ub_tab]) < len(unitary_block):
            comp_data.unitary_blocks[unitary_block_index] = [
                ONE_QUBIT_GATE_MAPPING[g](unitary_block[0].qubit)
                for g in comp_dict[ub_tab]
            ]

    # step 3: compile unitary blocks succeeding measurements
    for meas_gate_info, ub_info in comp_data.measurement_gates.items():
        unitaries_after = comp_data.unitary_blocks[ub_info["succeeding"]]
        updated_unitaries_after = _get_compilation_with_projectors_before_unitaries(
            comp_dict,
            unitaries_after,
            MEASUREMENT_GATE_MAPPING[meas_gate_info[2]],
            up_to_paulis=up_to_paulis,
        )
        comp_data.unitary_blocks[ub_info["succeeding"]] = [
            ONE_QUBIT_GATE_MAPPING[g](meas_gate_info[1].unique_identifier)
            for g in updated_unitaries_after
        ]

    # step 4: compile two qubit gates and surrounding unitaries
    two_qubit_dict_as_list = list(comp_data.two_qubit_gates.items())
    # define a convenience lambda for compiling the unitaries to natives,
    # for use in flipping in the two-qubit gates
    unitary_blocks_to_native: Callable[  # noqa: E731
        [Sequence[OneQubitCliffordGate]], list[OneQubitCliffordGate]
    ] = lambda ub: [
        ONE_QUBIT_GATE_MAPPING[g]
        for g in comp_dict[
            _get_tableau_key_from_sequence_of_gates(ub, up_to_paulis=up_to_paulis)
        ]
    ]
    for i in range(0, len(two_qubit_dict_as_list), 2):
        qubit1_gate_info, qubit1_unitary_info = two_qubit_dict_as_list[i]
        qubit2_gate_info, qubit2_unitary_info = two_qubit_dict_as_list[i + 1]

        ub1_before_reducing = [
            type(g) for g in comp_data.unitary_blocks[qubit1_unitary_info["preceding"]]
        ]
        ub2_before_reducing = [
            type(g) for g in comp_data.unitary_blocks[qubit1_unitary_info["succeeding"]]
        ]
        ub3_before_reducing = [
            type(g) for g in comp_data.unitary_blocks[qubit2_unitary_info["preceding"]]
        ]
        ub4_before_reducing = [
            type(g) for g in comp_data.unitary_blocks[qubit2_unitary_info["succeeding"]]
        ]

        two_qubit_gate_to_compile = TWO_QUBIT_GATE_MAPPING[qubit1_gate_info[2]]
        (
            ub1_after_reducing,
            ub2_after_reducing,
            ub3_after_reducing,
            ub4_after_reducing,
        ) = _get_compilation_with_two_qubit_gates(
            two_qubit_gate_to_compile,
            comp_dict,
            ub1_before_reducing,
            ub2_before_reducing,
            ub3_before_reducing,
            ub4_before_reducing,
            gate_exchange_dict,
            up_to_paulis=up_to_paulis,
            allow_terms_to_mutate=allow_terms_to_mutate,
            allow_terms_to_multiply=allow_terms_to_multiply,
        )
        if not issubclass(two_qubit_gate_to_compile, SymmetricTwoQubitGate):
            # If non-symmetric two-qubit gate, try flipping the orientation
            # of the gate to get a better compilation. This is achieved by
            # swapping the positions of the unitary blocks. We can do this
            # with one-qubit gates, using the CZ compilation of the 2q gate.

            # Note that CXSWAP is the only non-symmetric gate from the iSWAP
            # class, at least in stim currently.
            if two_qubit_gate_to_compile == CXSWAP:
                gate_to_intm_dict = GATE_TO_CZSWAP_DICT
                intm_to_target_dict = CZSWAP_TO_GATE_DICT
            else:
                gate_to_intm_dict = GATE_TO_CZ_DICT
                intm_to_target_dict = CZ_TO_GATE_DICT

            # get the unitaries used for compiling to and from CZ, but we will
            # swap the orientation of the adj. gates to swap the orientation
            # of the two-qubit gate
            # the unitaries in `intm_to_target_dict`` are guaranteed to be the adjoint
            # of those in `gate_to_intm_dict`` by construction
            ub1, ub2, ub3, ub4 = gate_to_intm_dict[two_qubit_gate_to_compile]
            ub1_adj, ub2_adj, ub3_adj, ub4_adj = intm_to_target_dict[
                two_qubit_gate_to_compile
            ]
            try:
                # If G_12 is the native (non-symmetric) 2-qubit gate, and G_21 is its
                # flipped version, then since G = u_1 \otimes u_3 CZ_12 u_2 \otimes u_4,
                # we can calculate that G = (u_1 \otimes u_3) (u_3adj \otimes u_1adj) CZ_12
                # (u_4adj \otimes u_2adj) (u_2 \otimes u_4).
                # this might produce a KeyError if the gates for swapping orientation
                # do not have a native gate expression
                ub3_with_flip_unitaries = unitary_blocks_to_native(
                    ub3_before_reducing + ub3 + ub1_adj
                )
                ub4_with_flip_unitaries = unitary_blocks_to_native(
                    ub2_adj + ub4 + ub4_before_reducing
                )
                ub1_with_flip_unitaries = unitary_blocks_to_native(
                    ub1_before_reducing + ub1 + ub3_adj
                )
                ub2_with_flip_unitaries = unitary_blocks_to_native(
                    ub4_adj + ub2 + ub2_before_reducing
                )
                flipping_unitaries_available = True
            except KeyError:
                # in this case, we cannot change the orientation of the gate,
                # since the unitaries to do so do not have an equivalent
                # expression in the provided native gates, so skip that part.
                flipping_unitaries_available = False
            if flipping_unitaries_available:
                (
                    ub3_after_reducing_flipped,
                    ub4_after_reducing_flipped,
                    ub1_after_reducing_flipped,
                    ub2_after_reducing_flipped,
                ) = _get_compilation_with_two_qubit_gates(
                    two_qubit_gate_to_compile,
                    comp_dict,
                    ub3_with_flip_unitaries,
                    ub4_with_flip_unitaries,
                    ub1_with_flip_unitaries,
                    ub2_with_flip_unitaries,
                    gate_exchange_dict,
                    up_to_paulis=up_to_paulis,
                    allow_terms_to_mutate=allow_terms_to_mutate,
                    allow_terms_to_multiply=allow_terms_to_multiply,
                )
                # if flipping gave a better compilation, reverse the direction
                # of the gate
                if len(ub1_after_reducing_flipped) + len(
                    ub3_after_reducing_flipped
                ) < len(ub1_after_reducing) + len(ub3_after_reducing):
                    (
                        ub1_after_reducing,
                        ub2_after_reducing,
                        ub3_after_reducing,
                        ub4_after_reducing,
                    ) = (
                        ub1_after_reducing_flipped,
                        ub2_after_reducing_flipped,
                        ub3_after_reducing_flipped,
                        ub4_after_reducing_flipped,
                    )
                    del comp_data.two_qubit_gates[qubit1_gate_info]
                    del comp_data.two_qubit_gates[qubit2_gate_info]
                    comp_data.two_qubit_gates[qubit2_gate_info] = qubit2_unitary_info
                    comp_data.two_qubit_gates[qubit1_gate_info] = qubit1_unitary_info

        # only ub2 or ub4 can potentially be longer, as we cannot add gates to ub1 or ub3
        delta = max(
            len(ub2_after_reducing) - len(ub2_before_reducing),
            len(ub4_after_reducing) - len(ub4_before_reducing),
        )
        if delta > 0:
            for i in range(qubit1_gate_info[0] + 1, max(layer_index_lookup.keys()) + 1):
                layer_index_lookup[i] += delta

        comp_data.unitary_blocks[qubit1_unitary_info["preceding"]] = [
            g(qubit1_gate_info[1].unique_identifier) for g in ub1_after_reducing
        ]
        comp_data.unitary_blocks[qubit1_unitary_info["succeeding"]] = [
            g(qubit1_gate_info[1].unique_identifier) for g in ub2_after_reducing
        ]
        comp_data.unitary_blocks[qubit2_unitary_info["preceding"]] = [
            g(qubit2_gate_info[1].unique_identifier) for g in ub3_after_reducing
        ]
        comp_data.unitary_blocks[qubit2_unitary_info["succeeding"]] = [
            g(qubit2_gate_info[1].unique_identifier) for g in ub4_after_reducing
        ]

    # step 5: compile unitaries before measurements
    for meas_gate_info, meas_gate_ub_info in comp_data.measurement_gates.items():
        updated_ub = _get_compilation_with_measurement_after_unitaries(
            comp_dict,
            comp_data.unitary_blocks[meas_gate_ub_info["preceding"]],
            MEASUREMENT_GATE_MAPPING[meas_gate_info[2]],
            up_to_paulis=up_to_paulis,
        )
        comp_data.unitary_blocks[meas_gate_ub_info["preceding"]] = [
            ONE_QUBIT_GATE_MAPPING[g](meas_gate_info[1].unique_identifier)
            for g in updated_ub
        ]

    # step 6: compile unitaries before and after resets
    for reset_gate_info, reset_gate_ub_info in comp_data.reset_gates.items():
        comp_data.unitary_blocks[reset_gate_ub_info["preceding"]] = []
        unitaries_after = comp_data.unitary_blocks[reset_gate_ub_info["succeeding"]]
        updated_unitaries_after = _get_compilation_with_projectors_before_unitaries(
            comp_dict,
            unitaries_after,
            RESET_GATE_MAPPING[reset_gate_info[2]],
            up_to_paulis=up_to_paulis,
        )
        comp_data.unitary_blocks[reset_gate_ub_info["succeeding"]] = [
            ONE_QUBIT_GATE_MAPPING[g](reset_gate_info[1].unique_identifier)
            for g in updated_unitaries_after
        ]

    return _create_circuit_from_compilation_data(
        comp_data, layer_index_lookup, iterations=comp_data.iterations
    )


def compile_circuit_to_native_gates(
    circuit: Circuit,
    native_gate_set: NativeGateSetAndTimes,
    up_to_paulis: bool = False,
    allow_terms_to_mutate: bool = True,
    allow_terms_to_multiply: bool = True,
) -> Circuit:
    """
    Compile a deltakit_circuit.Circuit to native gates and reduce the single qubit gate count
    where possible.

    Params
    ------
    circuit : Circuit
        deltakit_circuit.Circuit to compile.
    native_gate_set : NativeGateSetAndTimes
        Native gate set to compile to.
    up_to_paulis: bool, optional
        Specify whether compilation should happen only up to Pauli terms.
        By default, False.
    allow_terms_to_mutate : bool, optional
        For two qubit gate compilation, will allow the case of pulling a single-qubit Pauli
        term through a two-qubit gate such that the resulting (possibly two-qubit) Pauli has
        a different term on either qubit.
        E.g, pulling an X through a CZ from the first qubit mutates into a Z term
        on the second qubit.
        By default, True.
    allow_terms_to_multiply : bool, optional
        For two qubit gate compilation, will allow the case of pulling a term through a
        two-qubit gate such that a second term on the other qubit is created. E.g, pulling
        an X through a CX from the first qubit mutates into an X on the first qubit and an
        X on the second qubit.
        By default, True.

    Returns
    -------
    Circuit
        Compiled circuit.
    """
    # get compilation dictionary and gate exchange dict
    comp_dict, gate_exchange_dict = _get_compilation_dict(
        native_gate_set, up_to_paulis=up_to_paulis
    )

    # get CompilationData form of circuit
    comp_data = _extract_structure_from_circuit(circuit)
    layer_index_lookup = {i: i for i in range(comp_data.num_layers)}

    return merge_layers(
        _compile_comp_data(
            comp_data,
            layer_index_lookup,
            native_gate_set,
            comp_dict,
            gate_exchange_dict,
            up_to_paulis,
            allow_terms_to_mutate,
            allow_terms_to_multiply,
        )
    )



# (c) Copyright Riverlane 2020-2025.
"""This module contains an implementation of a data string wrapper,
which allows to send arbitrary data inside a request, without creating files.
"""
from __future__ import annotations

from pathlib import Path


class DataString:
    """Data string is a normal string which encodes arbitrary binary data (e.g.
    Stim circuit file, or measurements data).
    These strings may be used instead of file names in any calls to QEC
    Explorer client. The server recognises these strings by detecting a prefix,
    and treats them as data.

    `DataString` class provides methods to generate such string from files,
    strings and byte strings, as well as methods to parse them when the server responds
    with a data string.

    To generate a data string from a file::

        dstring_object = DataString.from_file(path_to_file)
        dstring = str(dstring_object)  # generate a data string
        len(dstring_object.data)  # raw data is stored in `DataString.data` field

    To parse a data string::

        dstring_object = DataString.from_data_string(datastring)

    To generate a data string from a variable::

        bytes_content = b"1,1,0,1,0\\n1,1,1,0,1\\n"
        csv_measurements = DataString(bytes_content)
        b8_measurements = qec_client.convert_format(
            input_file=csv_measurements,
            input_format="csv",
            output_format="b8",
        )
        b8_measurements.to_file("data.b8")

    To convert the content of a DataString object to a string::

        string = data_string.to_string()

    Data strings result in additional processing on client and server side,
    they introduce redundancy in data, and they also increase network traffic
    between client and server nodes.
    """

    empty: str = "duck://"
    """Empty data string"""

    def __init__(self, data: bytes | str = b"") -> None:
        """Construct an object from the data.

        Args:
            data (bytes | str, optional):
                bytes data to be encode, or a string, interpreted as
                UTF-8. Defaults to b"".
        """
        if isinstance(data, bytes):
            self._data = data
        elif isinstance(data, str):
            self._data = data.encode("utf-8")

    @staticmethod
    def from_file(filename: str | Path) -> DataString:
        """Create a data string from a file name.

        Args:
            filename (str | Path): file with data.

        Returns:
            DataString: object representing the data.
        """
        if isinstance(filename, str):
            filename = Path(filename)
        with filename.open("rb") as file:
            return DataString(file.read())

    @staticmethod
    def _hex_string_to_bytes(hex_string: str) -> bytes:
        return bytes.fromhex(hex_string)

    @staticmethod
    def is_data_string(data_string: str) -> bool:
        """Check that the string corresponds to
        a data string syntax.

        Args:
            data_string (str): string to check.

        Returns:
            bool: True if the string is a good data string.
        """
        if len(data_string) % 2 == 0:
            return False
        if not data_string.startswith(DataString.empty):
            return False
        return not set(data_string[7:]) - set("0123456789abcdef")

    @classmethod
    def from_data_string(cls, data_string: str) -> DataString:
        """Parse the data string into a bytes object.

        Args:
            data_string (str): string representation.

        Raises:
            ValueError:
                if the string is not a valid data string,
                error is raised.

        Returns:
            DataString: data string object.
        """
        if not cls.is_data_string(data_string):
            msg = (
                "String is not a valid Deltakit Explorer data string."
                " Expected that string starts with `duck://` and only "
                "contains characters `0123456789abcdef`"
            )
            raise ValueError(
                msg
            )
        return DataString(cls._hex_string_to_bytes(data_string[7:]))

    @property
    def data(self) -> bytes:
        """Read-only accessor to the data."""
        return self._data

    def to_file(self, path: str | Path) -> None:
        """Dump data to a file.

        Args:
            path (str | Path): file to save the data.
        """
        with Path.open(Path(path), "wb") as file:
            file.write(self._data)

    def to_string(self, encoding: str = "utf-8") -> str:
        """Converts DataString object to string.
        Args:
            encoding (str): The encoding with which to decode the bytes.
            Defaults to UTF-8.
        """
        return self._data.decode(encoding)

    def __str__(self) -> str:
        return self.empty + self._data.hex()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, DataString):
            return self._data == other.data
        return False

    def __hash__(self) -> int:
        return self._data.__hash__()

    def __repr__(self) -> str:
        return f"<DataString, data length={len(self._data)}>"
